import shutil

import numpy as np
import pytest

from besos.evaluator import EvaluatorGeneric
from besos.optimizer import rbf_opt
from besos.parameters import Parameter, RangeParameter
from besos.problem import Problem


@pytest.mark.skipif(shutil.which("bonmin") is None, reason="Requires bonmin to run.")
def test_rbfopt():
    """Simple test to make sure RBFOpt can be run properly"""

    def objective_function(x):
        return (x[0] * x[1] - x[2],)

    param_list = [Parameter(value_descriptors=RangeParameter(0, 10)) for _ in range(3)]
    problem = Problem(param_list, 1)

    evaluator = EvaluatorGeneric(objective_function, problem, error_mode="Silent")

    opt = rbf_opt(evaluator, 30)
    value = opt.iloc[0]

    print(opt)

    assert (
        np.isclose(value["RangeParameter [0, 10]"], 0)
        and (
            np.isclose(value["RangeParameter [0, 10]_1"], 5.062882681226724)
            or np.isclose(value["RangeParameter [0, 10]_1"], 0.5148690)
        )
        and np.isclose(value["RangeParameter [0, 10]_2"], 10)
        and np.isclose(value["outputs_0"], -10)
    ), f"Unexpected output: {value}"  # None_1 assert values hardcoded to pass on Debian



""" Helper functions to be used throughout the code base """


def listify(arg) -> list:
    """Converts the argument to a list if possible.
    If it is not iterable (i.e. one item), wrap it in a list.
    Treat strings as 1 item.
    :param arg: item(s) to be wrapped in list
    :return: List of the file arg(s)"""
    if isinstance(arg, str):
        return [arg]
    try:
        return list(arg)
    except TypeError:
        return [arg]


def get_file_name(path):
    """Finds and returns the file name from the path

    :param path: A directory to the idf file
    :return: String of file name
    """
    path = str(path).replace("\\", "/")
    if path.endswith("/"):
        path = path[:-2]
    file_name = path.split("/")[-1]
    return file_name



import shutil
import subprocess
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from unittest import mock as mocker

from besos import config
from besos import eppy_funcs as ef
from besos import pyehub_funcs as pf
from besos import sampling
from besos.evaluator import EvaluatorEH, EvaluatorEP, EvaluatorGeneric
from besos.parameters import (
    FieldSelector,
    Parameter,
    ParameterEH,
    PathSelector,
    RangeParameter,
)
from besos.problem import EHProblem, EPProblem, Problem


needs_glpk = pytest.mark.skipif(
    shutil.which("glpsol") is None,
    reason=("Requires a PuLP compatible solver. Tests assume GLPK is the solver used."),
)


@pytest.fixture
def building():
    # returns the basic building
    return ef.get_building()


@pytest.fixture
def unexpanded():
    return ef.get_building(str(Path(config.data_dir, "unexpanded.idf")))


@pytest.fixture
def hub():
    # returns the basic hub
    return pf.get_hub()


@pytest.fixture
def problem():
    parameters = [
        Parameter(
            FieldSelector(
                object_name="Mass NonRes Wall Insulation", field_name="Thickness"
            )
        )
    ]
    objectives = [
        "Electricity:Facility",
        "Gas:Facility",
    ]  # the default is just 'Electricity:Facility'

    return EPProblem(parameters, objectives)


@pytest.fixture()
def sample_ready_problem():
    parameters = [
        Parameter(
            FieldSelector(
                object_name="Mass NonRes Wall Insulation", field_name="Thickness"
            ),
            RangeParameter(0.1, 0.9),
        )
    ]
    objectives = [
        "Electricity:Facility",
        "Gas:Facility",
    ]  # the default is just 'Electricity:Facility'

    return EPProblem(parameters, objectives)


@pytest.fixture()
def inputs(sample_ready_problem):
    return sampling.dist_sampler(sampling.lhs, sample_ready_problem, 2)


@pytest.fixture
def energyhub_df():
    return pd.DataFrame(np.array([[200, 600], [600, 200]]), columns=["p1", "p2"])


@pytest.fixture
def energyplus_df():
    return pd.DataFrame(np.array([[0.5], [0.8]]), columns=["p1"])


@pytest.fixture
def hub_problem():

    parameters = [
        Parameter(PathSelector(["LINEAR_CAPITAL_COSTS", "Boiler"])),
        Parameter(PathSelector(["LINEAR_CAPITAL_COSTS", "CHP"])),
    ]
    objectives = ["total_cost", "total_carbon"]

    return EHProblem(parameters, objectives)


@pytest.fixture
def oldstyle_hub_problem():
    parameters = []
    message = (
        "Use a :class:`PathSelector` and a :class:`Parameter` instead. "
        "These have been swapped in automatically."
    )
    with pytest.warns(DeprecationWarning, match=message):
        parameters.append(ParameterEH(["LINEAR_CAPITAL_COSTS", "Boiler"]))
    with pytest.warns(DeprecationWarning, match=message):
        parameters.append(ParameterEH(["LINEAR_CAPITAL_COSTS", "CHP"]))
    objectives = ["total_cost", "total_carbon"]

    return EHProblem(parameters, objectives)


@needs_glpk
@pytest.mark.compatibility
def test_equivalent_problems(hub_problem, oldstyle_hub_problem, hub):
    evaluator_new = EvaluatorEH(hub_problem, hub)
    evaluator_old = EvaluatorEH(oldstyle_hub_problem, hub)
    result_new = evaluator_new([200, 600])
    result_old = evaluator_old([200, 600])
    assert np.allclose(
        result_new, result_old
    ), "Hub compatibility transformation does not work"


@pytest.mark.slow
def test_evaluatorEP(building, problem, regtest):

    """To make sure EvaluatorEP can be initialised and works as intended"""

    evaluator = EvaluatorEP(problem, building)
    result = evaluator([0.5])  # run with thickness set to 0.5

    with regtest:
        print([f"{x:.5E}" for x in result])


@needs_glpk
def test_evaluatorEH_single(hub, hub_problem, regtest):
    """To make sure EvaluatorEH can be initialised and works as intended"""

    evaluator = EvaluatorEH(hub_problem, hub)
    result = evaluator([200, 600])

    with regtest:
        print([f"{x:.5E}" for x in result])


@needs_glpk
def test_evaluatorEH_df(hub, hub_problem, energyhub_df, regtest):
    """To make sure EvaluatorEH df_apply works as intended"""

    evaluator = EvaluatorEH(hub_problem, hub)
    result = evaluator.df_apply(energyhub_df)

    with regtest:
        print(result)


@needs_glpk
@pytest.mark.slow
def test_evaluatorEH_EP(hub, hub_problem, building, problem, regtest):
    """To make sure that base EvaluatorEP output can be used in an EvaluatorEH"""
    evaluatorEP = EvaluatorEP(problem, building)
    evaluatorEH = EvaluatorEH(hub_problem, hub)
    result = evaluatorEH(evaluatorEP([0.5]))

    with regtest:
        print([f"{x:.5E}" for x in result])


@needs_glpk
def test_evaluatorEH_TS(hub, regtest):
    """To make sure that EvaluatorEH can accept a time series as input"""
    timeseries = [
        {
            0: 4.0,
            1: 8.0,
            2: 6.0,
            3: 5.0,
            4: 7.0,
            5: 7.0,
            6: 7.0,
            7: 7.0,
            8: 7.0,
            9: 7.0,
            10: 7.0,
        }
    ]

    time_series_parameters = [Parameter(PathSelector(["LOADS", "Elec"]))]

    objectives = ["total_cost", "total_carbon"]
    time_series_problem = EHProblem(time_series_parameters, objectives)
    evaluatorEH = EvaluatorEH(time_series_problem, hub)
    result = evaluatorEH(timeseries)

    with regtest:
        print([f"{x:.5E}" for x in result])


@needs_glpk
def test_evaluatorEH_TS_df(hub, regtest):
    """To make sure that EvaluatorEH's df_apply can accept a dataframe of time series as input"""
    default_timeseries = [
        {
            0: 1.0,
            1: 4.0,
            2: 4.0,
            3: 4.0,
            4: 4.0,
            5: 4.0,
            6: 4.0,
            7: 4.0,
            8: 4.0,
            9: 4.0,
            10: 4.0,
        },
        {
            0: 20.0,
            1: 20.0,
            2: 20.0,
            3: 20.0,
            4: 20.0,
            5: 20.0,
            6: 20.0,
            7: 12.0,
            8: 12.0,
            9: 12.0,
            10: 12.0,
        },
    ]
    just_modified_heat = [
        {
            0: 1.0,
            1: 4.0,
            2: 4.0,
            3: 4.0,
            4: 4.0,
            5: 4.0,
            6: 4.0,
            7: 4.0,
            8: 4.0,
            9: 4.0,
            10: 4.0,
        },
        {
            0: 18.0,
            1: 18.0,
            2: 18.0,
            3: 18.0,
            4: 18.0,
            5: 18.0,
            6: 18.0,
            7: 16.0,
            8: 16.0,
            9: 16.0,
            10: 16.0,
        },
    ]
    just_modified_elec = [
        {
            0: 4.0,
            1: 8.0,
            2: 6.0,
            3: 5.0,
            4: 7.0,
            5: 7.0,
            6: 7.0,
            7: 7.0,
            8: 7.0,
            9: 7.0,
            10: 7.0,
        },
        {
            0: 20.0,
            1: 20.0,
            2: 20.0,
            3: 20.0,
            4: 20.0,
            5: 20.0,
            6: 20.0,
            7: 12.0,
            8: 12.0,
            9: 12.0,
            10: 12.0,
        },
    ]
    modified_timeseries = [
        {
            0: 4.0,
            1: 8.0,
            2: 6.0,
            3: 5.0,
            4: 7.0,
            5: 7.0,
            6: 7.0,
            7: 7.0,
            8: 7.0,
            9: 7.0,
            10: 7.0,
        },
        {
            0: 18.0,
            1: 18.0,
            2: 18.0,
            3: 18.0,
            4: 18.0,
            5: 18.0,
            6: 18.0,
            7: 16.0,
            8: 16.0,
            9: 16.0,
            10: 16.0,
        },
    ]
    timeseries_df = pd.DataFrame(
        np.array(
            [
                default_timeseries,
                just_modified_heat,
                just_modified_elec,
                modified_timeseries,
            ]
        ),
        columns=["p1", "p2"],
    )

    TSDFparameters = [
        Parameter(PathSelector(["LOADS", "Elec"])),
        Parameter(PathSelector(["LOADS", "Heat"])),
    ]
    TSDFobjectives = ["total_cost", "total_carbon"]
    TSDFproblem = EHProblem(TSDFparameters, TSDFobjectives)

    evaluatorEH = EvaluatorEH(TSDFproblem, hub)
    result = evaluatorEH.df_apply(timeseries_df)

    with regtest:
        print(result)


@needs_glpk
@pytest.mark.slow
def test_evaluatorEH_EP_df(hub, hub_problem, building, problem, energyplus_df, regtest):
    """To make sure that dataframe EvaluatorEP output can be used in an EvaluatorEH"""
    evaluatorEP = EvaluatorEP(problem, building)
    evaluatorEH = EvaluatorEH(hub_problem, hub)
    result = evaluatorEH.df_apply(evaluatorEP.df_apply(energyplus_df))

    with regtest:
        print(result)


def test_evaluatorGeneric(problem):
    """To make sure EvaluatorGeneric can be initialised and works as intended"""

    def function(values):
        return (values[0], values[0] ** 2)

    # this denotes a problem which takes 1 input, produces 2 outputs and no constraints.
    # The placeholder parameters/objectives will be generated automatically.
    new_problem = Problem(1, 2, 0)

    evaluator_1 = EvaluatorGeneric(function, problem)
    evaluator_2 = EvaluatorGeneric(function, new_problem)
    result_1 = evaluator_1([4])
    result_2 = evaluator_2([4])

    assert result_1 == (
        4,
        16,
    ), f"Unexpected result for EvaluatorGeneric with EPProblem, {result_1}"
    assert result_2 == (
        4,
        16,
    ), f"Unexpected result for EvaluatorGeneric with custom problem, {result_2}"


@pytest.mark.slow
def test_parallel_evaluatorEP(building, problem, energyplus_df, regtest):
    """To make sure EvaluatorEP works with multiple processes"""

    evaluator = EvaluatorEP(problem, building)
    result = evaluator.df_apply(energyplus_df, processes=2, keep_input=True)

    with regtest:
        print(result)


@pytest.mark.slow
def test_error_evaluatorEP(problem):
    """To make sure EvaluatorEP error handling works"""
    building = ef.get_building(config.files.get("bad_idf"))
    evaluator = EvaluatorEP(problem, building)
    with pytest.raises(subprocess.CalledProcessError):
        with pytest.warns(UserWarning, match="problematic values were:"):
            evaluator([0.5])  # run with thickness set to 0.5


@pytest.mark.slow
def test_keep_output_evaluatorEP(building, sample_ready_problem, inputs, tmpdir):
    """Verify that the keep_output flag saves output in the correct location."""
    temp_path = tmpdir.mkdir("tmp")
    evaluator = EvaluatorEP(sample_ready_problem, building, out_dir=temp_path)
    df = evaluator.df_apply(inputs, keep_dirs=True)
    for dir_ in df["output_dir"]:
        # verify that the file exists and is readable
        with open(Path(dir_, "eplusout.dxf"), "r") as f:
            f.readline()


@pytest.fixture
def old_style_evaluation_function():
    def f(values):
        return (1.0, 2.0), ()

    return f


@pytest.mark.compatibility
def test_separate_outputs_and_constraints(old_style_evaluation_function):
    """Verify that we can still use the old evaluation function return format.

    Old format: (output1, output2, ...), (constraint1, constraint2, ...)
    New format: (output1, output2, ..., constraint1, constraint2, ...)

    """

    evaluator = EvaluatorGeneric(
        old_style_evaluation_function, Problem(inputs=1, outputs=2, constraints=0)
    )
    with pytest.warns(
        FutureWarning,
        match="Evaluators have changed the format used for storing outputs and constraints",
    ):
        assert evaluator([1]) == (1.0, 2.0)


@pytest.mark.compatibility
def test_epw_file_rename(building):
    """Verify that calls that use epw_file still work, but produce a warning."""
    with pytest.warns(FutureWarning, match="epw_file has been deprecated"):
        evaluator = EvaluatorEP(
            problem=Problem(), building=building, epw_file=config.files["epw"]
        )
    assert evaluator.epw == config.files["epw"]


@pytest.mark.compatibility
def test_epw_duplicate_values(building):
    """Verify that calls using both epw and epw_file are rejected."""
    with pytest.raises(ValueError, match="epw_file and epw cannot be used together"):
        EvaluatorEP(problem=Problem(), building=building, epw="a", epw_file="b")


@pytest.mark.slow
def test_expand_objects(unexpanded):
    EvaluatorEP(problem=Problem(), building=unexpanded)([])


# TODO add tests for cache/multiprocessing


def test_cache():
    func = mocker.Mock(return_value=[5])
    evaluator = EvaluatorGeneric(
        func, Problem([Parameter(value_descriptors=RangeParameter(0, 1))])
    )
    evaluator([1])
    # this call should hit the cache, and not the mocked evaluation function
    evaluator([1])
    func.assert_called_once()



"""
Defines various constants and defaults used to configure Besos.
"""

# Python Core Libraries
import os
import sys
import warnings
from pathlib import Path

# BESOS Imports
from errors import ModeError


baseDir = Path(__file__).resolve().parents[0]
# where to store output from simulations, or None for a temporary directory
out_dir = None
# folder name to store output in.
out_wrapper = "BESOS_Output"
# folder name to store errors in.
error_wrapper = "BESOS_Errors"
# where to move output files from failed runs
err_dir = Path(os.getcwd(), error_wrapper)
# where to look for input files
data_dir = Path(baseDir, "data")
# whether to use 'idf' or 'json' format by default
energy_plus_mode = "json"


# Produces a dictionary in the form {'extension': '/path/from/root/file.extension'}
# it is assumed that 'file.extension' is in the directory `inDir`
# When a file of the type 'json', 'idf', 'idd', 'epw', 'xlsx' or 'schema' is needed,
# the default is taken from this dictionary
files = {
    key: str(Path(data_dir, name).resolve())
    for key, name in {
        "json": "example_building.epJSON",
        "schema": "Energy+.schema.epJSON",
        "idf": "example_idf.idf",
        # note that some places use EnergyPlus-#-#-#/Energyplus.idd as the default idd instead
        # FIXME: address the above comment
        "idd": "example_idd.idd",
        "epw": "example_epw.epw",
        "xlsx": "example_xlsx.xlsx",
        "bad_idf": "example_bad_idf.idf",
    }.items()
}

# alias json and epJSON to the same building
if "json" in files and "epJSON" in files:
    warnings.warn("two defaults are created for json buildings")
elif "epJSON" in files:
    files["json"] = files["epJSON"]
elif "json" in files:
    files["epJSON"] = files["json"]

# set files['building'] equal to the building for the default mode
if energy_plus_mode == "idf":
    files["building"] = files["idf"]
    files["data dict"] = files["idd"]
elif energy_plus_mode == "json":
    files["building"] = files["json"]
    files["data dict"] = files["schema"]
else:
    raise ModeError(energy_plus_mode)

# TODO: solver_settings came from the config.yaml file that was removed, could be cleaned up
solver_settings = {
    "input_file": "data/example_xlsx.xlsx",
    "output_file": "test.xlsx",
    "solver": {"name": "glpk", "options": {"mipgap": 0.05}},
    "BIG_M": 99999,
}

# the default min and max values for range_parameters
range_parameter = dict(min=0, max=float("inf"))

# the default objective(s) to extract from simulations
objectives = "Electricity:Facility"  # TODO: Make this only apply to EvaluatorEP, not all evaluators

# The default `class_name` of meters/variables
objective_meter_type = "Output:Meter"
objective_variable_type = "Output:Variable"

# default arguments for optimizers
optimizer = dict(
    evaluations=1000,
)



"""
Functions to interact with eppy and geomeppy.
Some functions also support equivalent operations for json buildings.
"""

# Python Core Imports
import contextlib
import json
import os
import random
import re
import sqlite3
import string
import warnings
from functools import partial
from pathlib import Path
from typing import Optional

# External Libraries
import eppy
import numpy as np
import pandas as pd
from eppy.bunch_subclass import BadEPFieldError
from shapely.geometry import LineString, Point, Polygon

# BESOS Imports
from besos import config
from besos import eplus_funcs as eplus
from besos.IDF_class import IDF, view_epJSON
from besos.errors import ModeError


def view_building(building):
    """Displays a matplotlib 3d plot of the building"""
    if get_mode(building) == "json":
        return view_epJSON(building)
    elif get_mode(building) == "idf":
        return building.view_model()


def merge_building(building_from, building_to):
    """Copy objects in building_from to building_to."""
    mode_from = get_mode(building_from)
    mode_to = get_mode(building_to)
    if mode_to == "idf" and mode_from == "idf":
        for attr in building_from.idfobjects:
            for idfobject in building_from.idfobjects[attr]:
                building_to.copyidfobject(idfobject)
    elif mode_to == "json" and mode_from == "json":
        for attr in building_from:
            if attr in building_to:
                building_to[attr].update(building_from[attr])
            else:
                building_to[attr] = building_from[attr]
    else:
        raise NotImplementedError


def convert_format(s: str, place, mode):
    """Converts energyPlus names to their Eppy equivalents

    :param s: name to convert
    :param place: the type of place that is being used 'field' or 'class'
    :param mode: whether to convert to idf or json formatting
    :return: the converted name
    """
    # TODO: Find an authoritative source for the naming conventions
    if mode == "idf":
        if place == "field":
            return s.replace(" ", "_").replace("-", "")
        if place == "class":
            return s.upper()
    if mode == "json":
        if place == "field":
            return s.replace(" ", "_").replace("-", "_").lower()
        if place == "class":
            # uses camel case, s.title() removes some of the capitals and does not work
            return s
    raise ModeError(message=f"no format defined for place:{place} and mode:{mode}")


def get_mode(building) -> str:
    """Determines whether building uses an idf or json format

    :param building: a building, either an IDF or a dictionary of json data
    :return: "idf" or "json"
    """
    if isinstance(building, IDF):
        return "idf"
    if isinstance(building, dict):
        return "json"
    raise ModeError(message=f"Cannot find a valid mode for {building}")


def get_idf(
    idf_file: str = config.files.get("idf"),
    idd_file: str = None,
    version=None,
    output_directory=config.out_dir,
    ep_path=None,
) -> IDF:
    """Uses eppy to read an idf file and generate the corresponding idf object"""
    if version is None:
        version = eplus.get_idf_version_pth(idf_file)
    if idd_file is None:
        # set iddname to default value if needed.
        _, ep_directory = eplus.get_ep_path(version, ep_path)
        default_idd = Path(ep_directory, "Energy+.idd").resolve()
        if IDF.iddname is None:
            IDF.setiddname(str(default_idd))
        elif Path(IDF.iddname).resolve() != default_idd:
            warnings.warn(
                f"idd is already set to: {IDF.iddname}. "
                f"It will NOT be changed to the default of: {default_idd}."
            )
    else:
        # Trying to change the idd file inside a program will cause an error
        # paths must be converted to strings before sending to eppy
        IDF.setiddname(str(idd_file))
    # TODO: Fix this rather than hiding it.
    # calling IDF causes a warning to appear, currently redirect_stdout hides this.
    with contextlib.redirect_stdout(None):
        idf = IDF(idf_file)
    # override the output location so I stop messing up
    idf.run = partial(idf.run, output_directory=output_directory)
    return idf


def get_building(
    building=None,
    data_dict=None,
    output_directory=config.out_dir,
    mode=None,
    version=None,
    ep_path=None,
):
    """Get building from directory"""
    if mode is None and building is None:
        building = config.files.get("building")
        version = eplus.get_idf_version_pth(building)
        mode = config.energy_plus_mode
    if mode is None:
        ending = Path(building).suffix
        if ending == ".idf":
            mode = "idf"
        elif ending == ".epJSON":
            mode = "json"
        else:
            raise ModeError(ending)
    if building is None:
        building = config.files[mode]
    if version is None:
        version = eplus.get_idf_version_pth(building)
    if mode == "idf":
        idf = get_idf(
            idf_file=building,
            version=version,
            idd_file=data_dict,
            output_directory=output_directory,
            ep_path=ep_path,
        )
        return idf
    elif mode == "json":
        if data_dict is not None:
            warnings.warn(
                "epJSON does not support data_dict in get_building, pass into run_building if running EnergyPlus with 'schema_file' argument."
            )
        with open(building) as f:
            return json.load(f)
    raise ModeError(mode)


def get_idfobject_from_name(
    idf: IDF, name: str
) -> Optional[eppy.bunch_subclass.EpBunch]:
    """Gets an object from the passed idf where it's name
    value is equal to the passed string, if none are found
    then this method returns None


    :param idf: The idf to find the object from.
    :param name: The string to find that is equal to the name field of the object.
    :return: the object from the idf with the matching name.
    """

    #: Scan all objects in all lists
    for idfobject_list in idf.idfobjects.values():
        for obj in idfobject_list:
            #: Try except block incase object does not
            #: have a .Name attribute
            try:
                # If the object has the name we are looking for return it.
                if obj.Name == name:
                    return obj
            # ignore objects with no Name
            # Note that the BadEPFieldError may not be possible
            except (BadEPFieldError,):
                pass
    #: If this is encountered then no match was found, so we return None.
    return None


def get_windows(building):
    mode = get_mode(building)
    if mode == "idf":
        return (
            (window.Name, window)
            for window in building.idfobjects["FENESTRATIONSURFACE:DETAILED"]
        )
    elif mode == "json":
        return building["FenestrationSurface:Detailed"].items()
    else:
        raise ModeError(mode)


def wwr_all(building, wwr: float, direction=None) -> None:
    """Sets the wwr for all walls that have a window.
    Will malfunction if there are multiple windows on one wall
    """
    mode = get_mode(building)
    windows = get_windows(building)
    if direction is None:
        for window_name, window in windows:
            if "door" in window_name:
                warnings.warn(
                    f"\nGlass door was detected: {window_name}.\n"
                    f"Door will be resized with wwr_all\n"
                )
            wwr_single(window, wwr, building, mode)
    else:
        values = get_window_range(windows, building, mode)
        # TODO called get_windows twice, hard coded solution
        windows = get_windows(building)
        for window_name, window in windows:
            warnings.warn(
                f"\nGlass door was detected: {window_name}.\n"
                f"Door will be resized with wwr_all\n"
            )
            wwr_single(window, wwr, building, mode, direction, values)


def get_window_range(windows, building, mode):
    """get max and min coordinates range of windows"""
    values = {"max_x": -999, "min_x": 999, "max_y": -999, "min_y": 999}
    for window_name, window in windows:
        # getting coordinates here
        coordinates = get_coordinates(window, building, mode)
        xs = coordinates["xs"]
        ys = coordinates["ys"]
        zs = coordinates["zs"]
        # store max and min values so that we can know the range of coordinates
        if max(xs) > values["max_x"]:
            values["max_x"] = max(xs)
        if min(xs) < values["min_x"]:
            values["min_x"] = min(xs)
        if max(ys) > values["max_y"]:
            values["max_y"] = max(ys)
        if min(ys) < values["min_y"]:
            values["min_y"] = min(ys)
    return values


def get_coordinates(window, building, mode):
    """Get coordinates of the xs, ys, and zs"""
    if mode == "idf":

        def coordinates(ax):
            return [window[f"Vertex_{n}_{ax.upper()}coordinate"] for n in range(1, 5)]

    elif mode == "json":

        def coordinates(ax):
            return [window[f"vertex_{n}_{ax.lower()}_coordinate"] for n in range(1, 5)]

    else:
        raise ModeError(mode)
    coordinate = {}
    coordinate.update(
        {"xs": coordinates("X"), "ys": coordinates("Y"), "zs": coordinates("Z")}
    )
    return coordinate


def set_vertex(idfObj, vertexNum: int, x: float = 0, y: float = 0, z: float = 0):
    """Sets a single vertex of the passed idfObject (idfObj)
    to the specified x,y and z coordinates."""
    for val, name in zip((x, y, z), "XYZ"):
        idfObj["Vertex_{}_{}coordinate".format(vertexNum, name)] = round(val, 2)


def one_window(building):
    """Removes some windows so that each wall has no more than one"""
    mode = get_mode(building)
    walls = set()
    to_remove = []
    windows = get_windows(building)
    for window_name, window in windows:
        if mode == "idf":
            wall_name = window.Building_Surface_Name
        elif mode == "json":
            wall_name = window["building_surface_name"]
        else:
            raise ModeError(mode)
        if wall_name in walls:
            to_remove.append((window_name, window))
        else:
            walls.add(wall_name)
    if mode == "idf":
        for window_name, window in to_remove:
            building.idfobjects["FENESTRATIONSURFACE:DETAILED"].remove(window)
    elif mode == "json":
        for window_name, window in to_remove:
            building["FenestrationSurface:Detailed"].pop(window_name)
    else:
        raise ModeError(mode)


def wwr_single(window, wwr: float, building, mode, direction=None, values=None):
    """Process a single `window` to have the window to wall ratio specified by `wwr`"""

    # will not work for some orientations
    # multiple windows on a single wall will break this
    coordinates = get_coordinates(window, building, mode)
    # getting coordinates here
    xs = coordinates["xs"]
    ys = coordinates["ys"]
    zs = coordinates["zs"]
    # check the alignments
    if max(ys) == min(ys):
        axis = "x"
        axs = xs
    elif max(xs) == min(xs):
        axis = "y"
        axs = ys
    else:
        raise ValueError("The window is not aligned with the x or y axes")
    # with direction specified or not
    if direction is not None:
        if check_direct(direction, xs, ys, values):
            set_wwr_single(window, wwr, axs, zs, axis)
    else:
        set_wwr_single(window, wwr, axs, zs, axis)


# TODO document xs and ys
def check_direct(direction: str, xs, ys, values) -> bool:
    """check if the window's direction is the same as the desired direction

    :param direction: the direction to compare with
    :param values: the dictionary of max and min range of all windows
    :return:
    """
    direction = direction.upper()
    if direction not in {"NORTH", "WEST", "EAST", "SOUTH"}:
        raise NameError("Direction should be either north, east, west, or south")
    if type(direction) != str:
        raise TypeError("Direction should be a string")
    direct = ""
    if max(ys) == min(ys):
        if max(ys) == values["max_y"]:
            direct = "NORTH"
        else:
            direct = "SOUTH"
    elif max(xs) == min(xs):
        if max(xs) == values["max_x"]:
            direct = "EAST"
        else:
            direct = "WEST"
    return direct == direction


# TODO: Document axis
def set_wwr_single(window, wwr: float, axs, zs, axis):
    """Set the single window's wwr

    :param window: the window to be modified
    :param wwr: the window to wall ratio of the result
    :param axs: the axis this window is aligned with
    :param zs: the z coordinate of the window
    """
    width = max(axs) - min(axs)
    scale_factor = np.sqrt(wwr)
    new_width = width * scale_factor
    height = max(zs) - min(zs)
    new_height = height * scale_factor

    start_width = (width - new_width) / 2
    end_width = start_width + new_width
    start_height = (height - new_height) / 2
    end_height = start_height + new_height
    # Maintains vertex order by mimicking the current order
    s = [0] * 4
    for vertex in range(0, 4):
        if zs[vertex] == max(zs):
            # vertex on the top
            if axs[vertex] == max(axs):
                # TOP RIGHT
                s[0] += 1
                set_vertex(window, vertex + 1, z=end_height, **{axis: end_width})
            else:
                # TOP LEFT
                s[1] += 1
                set_vertex(window, vertex + 1, z=end_height, **{axis: start_width})
        else:
            if axs[vertex] == max(axs):
                # BOTTOM RIGHT
                s[2] += 1
                set_vertex(window, vertex + 1, z=start_height, **{axis: end_width})
            else:
                # BOTTOM LEFT
                s[3] += 1
                set_vertex(window, vertex + 1, z=start_height, **{axis: start_width})
    assert s == [1] * 4, ("verticesS are wrong:", s)


def set_daylight_control(building, zone_name, distance, illuminance=500):
    """Set daylighting control to the biggest window of the zone.

    :param building: an idf object
    :param zone_name: name of the zone
    :param distance: the distance from the reference point to the window
    :param illuminance: illuminance setpoint at reference point
    """
    surfs = [
        s.Name
        for s in building.idfobjects["BUILDINGSURFACE:DETAILED"]
        if s.Surface_Type.upper() == "WALL" and s.Zone_Name == zone_name
    ]
    windows = [
        w
        for w in building.idfobjects["FENESTRATIONSURFACE:DETAILED"]
        if w.Surface_Type.upper() == "WINDOW" and w.Building_Surface_Name in surfs
    ]
    if not windows:
        raise ValueError(f"No window found in {zone_name}.")
    window = windows[0]
    max_area = window.area
    for w in windows[1:]:
        area = w.area
        if area > max_area:
            max_area = area
            window = w
    vertex = get_vertices(building, zone_name, window, distance)
    if vertex is None:
        raise ValueError(f"Unable to find a daylighting reference point.")
    building.newidfobject(
        "Daylighting:Controls".upper(),
        Name=f"{zone_name} Daylighting Control",
        Zone_Name=zone_name,
        Minimum_Input_Power_Fraction_for_Continuous_or_ContinuousOff_Dimming_Control=0.01,
        Minimum_Light_Output_Fraction_for_Continuous_or_ContinuousOff_Dimming_Control=0.01,
        Glare_Calculation_Daylighting_Reference_Point_Name=f"{zone_name} Daylighting Reference Point",
        Daylighting_Reference_Point_1_Name=f"{zone_name} Daylighting Reference Point",
        Illuminance_Setpoint_at_Reference_Point_1=illuminance,
    )
    building.newidfobject(
        "Daylighting:ReferencePoint".upper(),
        Name=f"{zone_name} Daylighting Reference Point",
        Zone_Name=zone_name,
        XCoordinate_of_Reference_Point=vertex[0],
        YCoordinate_of_Reference_Point=vertex[1],
        ZCoordinate_of_Reference_Point=vertex[2],
    )


def get_vertices(building, zone_name, window, distance):
    v1 = (
        window.Vertex_1_Xcoordinate,
        window.Vertex_1_Ycoordinate,
        window.Vertex_1_Zcoordinate,
    )
    v2 = (
        window.Vertex_2_Xcoordinate,
        window.Vertex_2_Ycoordinate,
        window.Vertex_2_Zcoordinate,
    )
    v3 = (
        window.Vertex_3_Xcoordinate,
        window.Vertex_3_Ycoordinate,
        window.Vertex_3_Zcoordinate,
    )
    vertices = [v1, v2, v3]

    def subtract(v1, v2):
        return v1[0] - v2[0], v1[1] - v2[1], v1[2] - v2[2]

    for i in range(3):
        for j in range(i + 1, 3):
            line = subtract(vertices[i], vertices[j])
            if line[2] == 0:
                point1, midpoint = (
                    vertices[j],
                    (
                        (vertices[j][0] + vertices[i][0]) / 2,
                        (vertices[j][1] + vertices[i][1]) / 2,
                    ),
                )
            elif line[1] == 0 or line[0] == 0:
                height = abs(line[2]) / 2
    r_point1, r_point2 = get_reference_point(point1[:2], midpoint, distance)
    r_point = point_in_zone(building, zone_name, [r_point1, r_point2])
    if r_point is not None:
        return r_point + (height,)
    return None


def point_in_zone(building, zone_name, points):
    surfs = building.idfobjects["BUILDINGSURFACE:DETAILED"]
    zone_surfs = [s for s in surfs if s.Zone_Name == zone_name]
    floors = [s for s in zone_surfs if s.Surface_Type.upper() == "FLOORS"]
    for floor in floors:
        for p in points:
            if point_in_surface(floor, p):
                return p
    roofs = [s for s in zone_surfs if s.Surface_Type.upper() in ["ROOF", "CEILING"]]
    for roof in roofs:
        for p in points:
            if point_in_surface(roof, p):
                return p
    return None


def point_in_surface(surface, p):
    v1 = (surface.Vertex_1_Xcoordinate, surface.Vertex_1_Ycoordinate)
    v2 = (surface.Vertex_2_Xcoordinate, surface.Vertex_2_Ycoordinate)
    v3 = (surface.Vertex_3_Xcoordinate, surface.Vertex_3_Ycoordinate)
    v4 = (surface.Vertex_4_Xcoordinate, surface.Vertex_4_Ycoordinate)
    poly = Polygon((v1, v2, v3, v4))
    point = Point(p)
    return poly.contains(point)


def get_reference_point(p1, p2, distance):
    line1 = LineString([p1, p2])
    left = line1.parallel_offset(distance, "left")
    right = line1.parallel_offset(distance, "right")
    p3 = left.boundary[1]
    p4 = right.boundary[0]
    return (p3.x, p3.y), (p4.x, p4.y)


def read_sql(path: str, cmds: list, direction=None):
    """Open sql file with connect

    :param path: absolute directory to the sql file.
    :param cmds: list of commands need to be processed, 'all' will process all cmds.
        All commands available : 'all', 'wall area', 'ceiling height', 'floor area', 'volume'
    :param direction: a list of directions of walls when process wall area cmd,
        None means taking both area and gross area of walls in all directions and
        the total areas of all walls.
    :return: a dictionary of all data desired.

    """
    try:
        sql_path = path
        conn = sqlite3.connect(sql_path)
        cur = conn.cursor()
    except IOError:
        print("Either no such file or file path is not correct.")

    if not isinstance(cmds, list):
        raise TypeError("cmds has be a list")

    def get_azim_range(direction: str):
        if direction == "total":
            azim_range = [0, 360]
        elif direction == "north":
            azim_range = [0, 44.99, 315, 360]
        elif direction == "east":
            azim_range = [45, 135]
        elif direction == "south":
            azim_range = [135, 269.99]
        elif direction == "west":
            azim_range = [270, 515]
        else:
            raise NameError(
                "Wrong direction! attention on the str, 'north', 'south', 'west', 'east'"
            )
        return azim_range

    def wall_area(dire: str):
        """Find total area with window subtracted from either all walls or only walls in current direction"""
        table_name = "Surfaces"
        azim_range = get_azim_range(dire)
        if len(azim_range) > 2:
            query = (
                f"SELECT SUM(Area), SUM(GrossArea) FROM {table_name} WHERE "
                f"(ClassName = 'Wall' AND ((Azimuth >= {azim_range[0]} AND "
                f"Azimuth <= {azim_range[1]}) or (Azimuth >= {azim_range[2]} "
                f"AND Azimuth <= {azim_range[3]})))"
            )
        else:
            query = (
                f"SELECT SUM(Area), SUM(GrossArea) FROM {table_name} WHERE "
                f"(ClassName = 'Wall' AND Azimuth >= {azim_range[0]} AND "
                f"Azimuth <= {azim_range[1]})"
            )
        cur.execute(query)
        data = cur.fetchall()
        return data

    def ceiling_height():
        """Find height of the ceiling"""
        floor = floor_area()
        volume = zone_volume()
        height = volume / floor
        return height

    def floor_area():
        """Find total floor area"""
        table_name = "Zones"
        query = f"SELECT SUM(FloorArea) FROM {table_name}"
        cur.execute(query)
        data = cur.fetchall()
        return data[0][0]

    def zone_volume():
        """Find total volume of Zones"""
        table_name = "Zones"
        query = f"SELECT SUM(Volume) FROM {table_name}"
        cur.execute(query)
        data = cur.fetchall()
        return data[0][0]

    def sg_temp():
        """find site ground temperature"""
        table_name = "ReportDataDictionary"
        query = f"SELECT ReportDataDictionaryIndex FROM {table_name} WHERE Name = 'Site Ground Temperature'"
        cur.execute(query)
        index = cur.fetchall()
        table_name = "ReportData"
        query = f"SELECT Value FROM {table_name} WHERE ReportDataDictionaryIndex = {index[0][0]}"
        cur.execute(query)
        data = cur.fetchall()
        return data

    # Taking commands, there is a small problem is that the code is not checking if the direction is correct.
    # And there is no protection when using cmd all and others
    def pull_data():
        result = {}
        # decide which direction we want
        if direction is None:
            direct = ["total", "south", "north", "west", "east"]
        else:
            direct = direction
        if len(cmds) == 1 and cmds[0] == "all":
            result.update(
                {
                    "ceiling height": ceiling_height(),
                    "floor area": floor_area(),
                    "volume": zone_volume(),
                }
            )
            for dire in direct:
                area = wall_area(dire)
                result.update(
                    {
                        f"{dire} wall area": area[0][0],
                        f"{dire} wall gross area": area[0][1],
                    }
                )
        else:
            for cmd in cmds:
                if cmd == "all":
                    raise NameError(
                        "There should not be an 'all' command with other commands"
                    )
                elif cmd == "wall area":
                    for dire in direct:
                        area = wall_area(dire)
                        result.update(
                            {
                                f"{dire} wall area": area[0][0],
                                f"{dire} wall gross area": area[0][1],
                            }
                        )
                elif cmd == "ceiling height":
                    result.update({"ceiling height": ceiling_height()})
                elif cmd == "floor area":
                    result.update({"floor area": floor_area()})
                elif cmd == "volume":
                    result.update({"volume": zone_volume()})
                elif cmd == "SGtemp":
                    result.update({"site ground temp": sg_temp()})
                else:
                    raise NameError(
                        "No such command, available cmds are 'all', 'wall area', "
                        "'ceiling height', 'floor area', 'volume', 'SGtemp'"
                    )
        return result

    result = pull_data()
    conn.close()
    return result


# TODO make this function customizable for more cmds, not just a handler for one task.
# shape option is for special use, might remove it in the future
# number of floor is also a temp solution
# Add direction
# Add cmds
def write_csv(path: str, dest: str, shape=False):
    """A handler to call read_sql() function and get data, then put the data into exl file

    :param path: path to read sql file
    :param dest: destination to put data
    :param shape: for special use, only works with right file name
    """
    # Get data from sql file. Currently retrieves all the data.
    data = read_sql(
        path,
        ["all"],
    )

    # Get number of floors and shape from file name
    filename = Path(dest).name
    match = re.search(r"M.idf", filename)
    if match is None:
        num_of_floor = 1
    else:
        num_of_floor = 3
    if shape:
        shape = filename[0].upper()

    # Read csv file
    df = pd.read_csv(dest)
    wall_area = []
    south_wall_area = []
    total_rows = len(df["Window to Wall Ratio"])
    for i in range(total_rows):
        total_area = data["total wall gross area"] * (1 - df["Window to Wall Ratio"][i])
        south_area = data["south wall gross area"] * (1 - df["Window to Wall Ratio"][i])
        wall_area.append(total_area)
        south_wall_area.append(south_area)

    # put data back to csv
    df["wall area net"] = wall_area
    df["wall area gross"] = data["total wall gross area"]
    df["south wall area net"] = south_wall_area
    df["south wall area gross"] = data["south wall gross area"]
    df["number of floors"] = num_of_floor
    df["ceiling height"] = data["ceiling height"]
    df["total floor area"] = data["floor area"]
    df["total volume"] = data["volume"]
    if shape:
        df["shape"] = shape
    df.to_csv(dest)


def generate_dir(dest_folder=None):
    """func use to generate a directory for besos outputs"""
    folder = "BESOS_Output"
    if dest_folder is not None:
        folder = dest_folder
    res = "".join(random.choices(string.ascii_uppercase + string.digits, k=20))
    while os.path.exists(Path(folder, res)):
        res = "".join(random.choices(string.ascii_uppercase + string.digits, k=20))
    os.makedirs(Path(folder, res))
    dir_ = Path(folder, res)
    return dir_


def generate_batch(
    account: str,
    time: str,
    email: str,
    task_id: int,
    cpu_per_task=1,
    mem=1000,
):
    """function to write a bash file for running jupyter on computer canada

    :param account: account used
    :param time: time for bash job
    :param email: user email
    :param task_id: task id to use
    :param cpu_per_task: number of cpu used to run one task
    :param mem: memory needs in total
    """
    # TODO there might be a calculation relationship between mem, cpu and task_id.
    # If possible give some default value to the arguments

    f = open("clusterbatch.sh", "w")
    f.write(
        f"#!/bin/bash\n"
        f"#SBATCH --account={account}\n"
        f"\n"
        f"#SBATCH --array=1-250\n"
        f"#SBATCH --time={time}\n"
        f"#SBATCH --cpus-per-task={cpu_per_task}\n"
        f"#SBATCH --mem={mem}mb\n"
        f"#SBATCH --mail-user={email}\n"
        f"\n"
        f"#!generate the virtual environment\n"
        f"module load python/3.6\n"
        f"source ~/env/bin/activate\n"
        f"echo 'program started at:`date`'\n"
        f"srun python cluster.py $SLURM_ARRAY_TASK_ID {task_id}\n"
        f"deactivate\n"
        f"echo 'program ended at: `date`'\n"
    )
    f.close()


def convert_to_json(idf: IDF):
    """convert idf file to json
    the func will create a json file that is converted from the idf
    can use --convert-only when energyplus version 9.3 is in besos
    """
    os.system(f"energyplus -c {idf}")



"""
This file contains custom errors used by BESOS
"""
import platform


class ModeError(ValueError):
    """An error for when an invalid mode is encountered."""

    def __init__(self, mode=None, message=None):  # pragma: no cover
        if message is None:
            message = f'Invalid mode {mode}. Expected "idf" or "json"'
        super().__init__(message)


class InstallationError(ValueError):
    """
    idf version X does not match with installed EP, please supply ep_path to correct EP version.
    We expect EP to be installed in {}.
    """

    def __init__(self, version=None, ep_path=None, message=None):
        if message is None and ep_path and version:
            message = f"EnergyPlus in IDF does not match ep_path EP version, please supply correct ep_path to v{version}"
            message += f"\n ep_path: {ep_path}\n Version: {version}"
        elif message is None and version:
            message = f"\n Could not locate EnergyPlus V{version}"
            version = version.replace(".", "-")

            if platform.system() == "Windows":
                ep_directory = f"Detected Windows OS: C:/EnergyPlusV{version}"
            elif platform.system() == "Linux":
                ep_directory = f"Detected Linux OS: /usr/local/EnergyPlus-{version}"
            else:
                ep_directory = f"Detected MacOS: /Applications/EnergyPlus-{version}"

            message += f"\n We expected EnergyPlus to be installed in the following path\n {ep_directory}\n Or specify in ep_path argument\n"

        super().__init__(message)



"""
Classes used to represent the attributes of the building that can be varied,
such as the thickness of the insulation, or the window to wall ratio.
These parameters are separate from the value that they take on during any evaluation of the model.
"""

# Python Core Libraries
import warnings
from abc import ABC, abstractmethod
from typing import Callable, Dict, Iterable, List, Tuple, Union

# External Libraries
import platypus
from deprecated.sphinx import deprecated

# BESOS Imports
from besos import eppy_funcs as ef
from besos import pyehub_funcs as pf
from besos import besos_utils
from besos.IO_Objects import AnyValue, Descriptor, DummySelector, Selector
from besos.IO_Objects import ReprMixin
from besos.config import range_parameter as conf
from besos.errors import ModeError


class RangeParameter(Descriptor):
    """Represents a value that is contained in an interval.

    .. warning:: RangeParameter is actually a kind of
        :class:`Descriptor <IO_Objects.Descriptor>`,
        not a :class:`Parameter`. This naming may be confusing.
    """

    pandas_type = float
    rbf_type = "R"

    def __init__(
        self,
        min_val: float = conf.get("min"),
        max_val: float = conf.get("max"),
        name="",
    ):
        """

        :param max_val: the minimum value
        :param min_val: the maximum value
        """
        super().__init__(name=name)
        if min_val > max_val:
            raise ValueError("minimum is larger than maximum")
        self.min = min_val
        self.max = max_val
        self._add_reprs(["min", "max"])
        self.platypus_type = platypus.Real(self.min, self.max)

    def validate(self, value: float) -> bool:
        """Checks if `value` is contained within the range described by this Descriptor

        :param value: The value to check.
        :return: True if the value in the range, False otherwise.
        """
        min_ = float("-inf") if self.min is None else self.min
        max_ = float("inf") if self.max is None else self.max
        return min_ <= value <= max_

    def sample(self, value: float) -> float:
        """Transforms a value in [0, 1] into a value in [min, max].
        This transformation is uniform.

        :param value: a value in the interval [0, 1]
        :return: a value from the interval [min, max]
        """
        return (self.max - self.min) * value + self.min

    def __str__(self):
        return f"{self.__class__.__name__} [{self.min}, {self.max}]"

    @property
    def _default_name(self):
        return str(self)


class _Category(platypus.Subset):
    def __init__(self, elements):
        super().__init__(elements=elements, size=1)

    def encode(self, value):
        return [value]

    def decode(self, value):
        return value[0]


class CategoryParameter(Descriptor):
    """Represents value that is selected from a list of possible values.

    .. warning:: CategoryParameter is actually a kind of
        :class:`Descriptor <IO_Objects.Descriptor>`,
        not a :class:`Parameter`.
        This naming may be confusing.
    """

    pandas_type = "category"

    def __init__(self, options: Iterable, **kwargs):
        """

        :param options: a list of possible value this parameter can be set to
        """
        super().__init__(**kwargs)
        self.options = list(options)
        self._add_repr("options")
        self.platypus_type = _Category(self.options)

    def validate(self, value):
        return value in self.options

    def sample(self, value):
        return self.options[int(len(self.options) * value)]

    # This is not a great default
    @property
    def _default_name(self):
        return f"Category{self.options}"


class AbstractFieldSelector(Selector, ABC):
    """Base class for selectors that modify one field in one or more objects in an EnergyPlus building"""

    def __init__(self, field_name):
        super().__init__()
        self.field_name = field_name
        self._add_repr("field_name")

    def get(self, building) -> List:
        """Gets the current values of this field from a building

        :param building: the building to retrieve values from
        :return: a list containing the current values of this selector's fields
        """
        mode = ef.get_mode(building)
        objects = self.get_objects(building)
        field_name = ef.convert_format(self.field_name, "field", mode)
        if mode == "idf":
            return [getattr(o, field_name) for o in objects]
        if mode == "json":
            return [o[field_name] for o in objects]

    def set(self, building, value) -> None:
        """Sets this field in the building to the provided value

        :param building: the building to modify
        :param value: the value to set this field to
        :return:
        """
        mode = ef.get_mode(building)
        objects = self.get_objects(building)
        field_name = ef.convert_format(self.field_name, "field", mode)
        if mode == "idf":
            for o in objects:
                setattr(o, field_name, value)
        if mode == "json":
            for o in objects:
                assert field_name in o, f"{field_name} not in {repr(o)}"
                o[field_name] = value

    def multiply(self, building, value) -> None:
        """Multiplies this field in the building by the provided value

        :param building: the building to modify
        :param value: the value to multiply this field by
        :return:
        """
        mode = ef.get_mode(building)
        objects = self.get_objects(building)
        field_name = ef.convert_format(self.field_name, "field", mode)
        if mode == "idf":
            for o in objects:
                setattr(o, field_name, value * getattr(o, field_name))
        if mode == "json":
            for o in objects:
                assert field_name in o, f"{field_name} not in {repr(o)}"
                o[field_name] = value * o[field_name]

    def add(self, building, value) -> None:
        """Adds to this field in the building by the provided value

        :param building: the building to modify
        :param value: the value to add to this field
        :return:
        """
        mode = ef.get_mode(building)
        objects = self.get_objects(building)
        field_name = ef.convert_format(self.field_name, "field", mode)
        if mode == "idf":
            for o in objects:
                setattr(o, field_name, value + getattr(o, field_name))
        if mode == "json":
            for o in objects:
                assert field_name in o, f"{field_name} not in {repr(o)}"
                o[field_name] = value + o[field_name]

    @abstractmethod
    def get_objects(self, building) -> List:
        """Returns a list of the object this selector applies to

        :param building: the building to search for objects
        :return: a list of the objects this selector applies to
        """
        pass


class FilterSelector(AbstractFieldSelector):
    """A selector that uses a custom function to find which objects it should modify"""

    def __init__(self, get_objects, field_name):
        """

        :param get_objects: a function that takes a building and returns the objects this selector should modify
        :param field_name: the field to modify
        """
        super().__init__(field_name)
        self._get_objects = get_objects
        self._add_repr("get_objects", "_get_objects")

    def get_objects(self, building):
        return self._get_objects(building)


class FieldSelector(AbstractFieldSelector):
    """A selector that modifies one or more fields in an EnergyPlus building,
    based on the class, object and field names"""

    def __init__(
        self, class_name: str = None, object_name: str = None, field_name: str = None
    ):
        """

        :param class_name: class of the object to modify
            ex: 'Material'
        :param object_name: name of the object to modify
            ex: 'Mass NonRes Wall Insulation'
        :param field_name: name of the field to modify
            ex: Thickness
        """
        super().__init__(field_name=field_name)
        self.class_name = class_name
        self.object_name = object_name
        self._add_reprs(["class_name", "object_name"], check=True)

    def get_objects(self, building) -> List:
        """Retrieves the objects that this selector will affect from the building.

        :param building: the building to search
        :return: a list of the objects found
        """
        mode = ef.get_mode(building)
        if mode == "idf":
            if self.class_name is not None:
                class_name = ef.convert_format(self.class_name, "class", "idf")
            else:
                class_name = None
            if self.object_name == "*":
                if class_name is None:
                    raise TypeError(
                        "When object_name='*', class_name must be specified."
                    )
                return building.idfobjects[class_name]
            if self.object_name and class_name:
                # this is probably the most reliable way to select an idfObject.
                return [building.getobject(key=class_name, name=self.object_name)]
            if self.object_name:
                # There should only one object matching the name, assuming the idf is valid
                return [ef.get_idfobject_from_name(building, self.object_name)]
            if class_name is not None:
                # assume that we want the first object matching the key
                # TODO: is this specific enough, or should we remove it? our JSON code does not support this
                return [building.idfobjects[class_name][0]]
            else:  # we have neither object_name nor class_name
                raise TypeError("Either class_name or object_name must be specified.")
        elif mode == "json":
            if self.object_name == "*":
                if not self.class_name:
                    raise TypeError(
                        "When object_name='*', class_name must be specified."
                    )
                return list(building[self.class_name].values())
            if self.object_name and self.class_name:
                # this is probably the most reliable way to select an idfObject.
                return [building[self.class_name][self.object_name]]
            if self.object_name:
                # There should only one object matching the name, assuming the building is valid
                result = [
                    obj
                    for objs in building.values()
                    for name, obj in objs.items()
                    if name == self.object_name
                ]
                if len(result) != 1:
                    warnings.warn(
                        f"found {len(result)} objects with object_name: {self.object_name}, expected 1"
                    )
                return result
            if self.class_name:
                result = list(building[self.class_name].items())
                if len(result) == 1:
                    return result
                raise ValueError(
                    f"multiple objects with class_name {self.class_name}."
                    f"Cannot guarantee a reliable ordering"
                )
            else:  # we have neither object_name nor class_name
                raise TypeError("Either class_name or object_name must be specified.")
        raise ModeError(mode)


class PathSelector(Selector):
    """A Selector for modifying EnergyHub objects using a path."""

    def __init__(self, parameter_path=None):
        """
        :param parameter_path: the path to the parameter to modify for the EnergyHub
        """
        super().__init__()
        self.parameter_path = parameter_path

    def get(self, hub):
        return pf.get_by_path(hub.__dict__, self.parameter_path)

    def set(self, hub, value):
        pf.set_by_path(hub.__dict__, self.parameter_path, value)

    def multiply(self, hub, value):
        self.set(hub, value * self.get(hub))

    def add(self, hub, value):
        self.set(hub, value + self.get(hub))

    def setup(self, hub) -> None:
        """Modifies the building so that it is ready for this selector"""
        pass


class GenericSelector(Selector):
    """A selector that supports custom get/set functions"""

    def __init__(
        self, set: Callable = None, get: Callable = None, setup: Callable = None
    ):
        """

        :param set: The function to use when setting.
            Must accept a building and a value. Can modify the building in any way.
        :param get: The function to use when getting a value from a building.
            Must accept a building, and should return a list of the current values
            of the fields this selector affects.
        :param setup: The function to use when setting up the building.
            This function must accept a building. It may modify the building in any way.
            This will be run once when the evaluator is initialized with a building.
            (or when the building the evaluator uses is changed.)
        """
        super().__init__()
        self._set = set
        self._get = get
        self._setup = setup
        for attr in ("set", "get", "setup"):
            self._add_repr(attr, f"_{attr}", True)

    def set(self, building, *values):
        """Redirects to the custom function given. The building and values are passed to the function.

        :param building: the building to modify
        :param values: the values to use when modifying the building
        """
        if self._set is not None:
            return self._set(building, *values)
        raise NotImplementedError

    def get(self, building):
        """A custom function for getting the current values of this field from a building

        :param building: the building to retrieve values from
        """
        if self._get is not None:
            return self._get(building)
        raise NotImplementedError

    def setup(self, building):
        """A custom function for modifying the building so that it is ready for this selector

        :param building: the building to modify
        """
        if self._setup is None:
            return
        self._setup(building)


# Parameters


class Parameter(ReprMixin):
    def __init__(
        self,
        selector: Selector = None,
        value_descriptors: Union[Descriptor, List[Descriptor]] = None,
        name="",
        *,
        value_descriptor: Descriptor = None,
    ):
        """

        :param selector: a Selector describing how to modify the building
        :param value_descriptors: a Descriptor specifying which values to use
        :param name: the name of the Parameter. Used for readability
            and column labelling.
        """
        super().__init__()
        self.selector = selector or DummySelector()

        # handle old call format
        if value_descriptor is not None and value_descriptors is not None:
            raise ValueError("Cannot use both value_descriptor and value_descriptors")
        elif value_descriptor:
            warnings.warn(
                FutureWarning("Use value_descriptors instead of value_descriptor.")
            )
            value_descriptors = value_descriptor

        # assign default value
        if value_descriptors is None:
            # no inputs is probably not useful, but an empty list allows it
            value_descriptors = [AnyValue()]
        self.value_descriptors = besos_utils.listify(value_descriptors)

        # handle name as input to a parameter to support old call signature
        if name:
            descriptor = self.value_descriptor
            if descriptor.has_default_name():
                descriptor.name = name
            else:
                warnings.warn(
                    UserWarning(
                        f"This parameter's descriptor is already named "
                        f"{descriptor.name}."
                        f"The name used as an input ({name}) will be discarded."
                    )
                )

        self._add_reprs(["selector", "value_descriptors"], True)

    def transformation_function(self, building, value_dict) -> None:
        """Mutates the building based on the value provided."""
        relevant_values = (
            value_dict[descriptor] for descriptor in self.value_descriptors
        )
        self.selector.set(building, *relevant_values)

    @deprecated(
        version="2.0.0",
        reason="Does not support multiple descriptors per parameter."
        "Use the sample method of the this Parameter's Descriptor(s) instead.",
    )
    def sample(self, value: float):
        """Takes a value in the range 0-1 and returns a valid value for this parameter"""
        return self.value_descriptor.sample(value)

    @deprecated(
        version="2.0.0",
        reason="Does not support multiple Descriptors per Parameter."
        "Use the validate method of this Parameter's Descriptor(s) instead.",
    )
    def validate(self, value):
        """Checks if value is a valid value for this parameter.

        :param value:
        :return: True if the value is valid False otherwise
        """
        return self.value_descriptor.validate(value)

    def setup(self, building) -> None:
        self.selector.setup(building)

    @property
    @deprecated(
        version="2.0.0",
        reason="Parameters are no longer nameable. "
        "Use the _default_name(s) or name(s) "
        "of this Parameter's Descriptor(s) instead",
    )
    def _default_name(self) -> str:
        """The name to use for this Parameter if no name was provided"""
        if hasattr(self.selector, "field_name"):
            return self.selector.field_name
        else:
            warnings.warn(UserWarning, f"No name found for parameter {repr(self)}")
            return "Unnamed"

    @property
    @deprecated(
        version="2.0.0",
        reason="Does not support multiple Descriptors per Parameter."
        "Use the platypus_type of this Parameter's Descriptor(s) instead.",
    )
    def platypus_type(self):
        """The platypus equivalent of this parameter"""
        return self.value_descriptor.platypus_type

    @property
    @deprecated(
        version="2.0.0",
        reason="Does not support multiple Descriptors per Parameter."
        "Use the value_descriptors of this Parameter instead.",
    )
    def value_descriptor(self) -> Descriptor:
        num = len(self.value_descriptors)
        if num == 1:
            return self.value_descriptors[0]
        raise ValueError(
            f"this parameter has {num} value descriptors."
            "You are trying to use an outdated feature that"
            "assumes there is exactly one descriptor."
        )

    @property
    @deprecated(
        version="2.0.0",
        reason="Parameters are no longer nameable. "
        "Use the name(s) of this Parameter's Descriptor(s) instead",
    )
    def name(self) -> str:
        return self.value_descriptor.name


@deprecated(
    version="1.6.0",
    reason=(
        "Use a :class:`PathSelector` and a :class:`Parameter` instead. "
        "These have been swapped in automatically."
    ),
)
class ParameterEH:
    """A parameter for use with EnergyHub"""

    # this method overrides the default behaviour when trying to create a new ParameterEH
    # instead of creating a ParameterEH, it creates a PathSelector,
    # puts it in a Parameter, and returns that parameter.
    def __new__(
        cls,
        parameter_path: List[str] = None,
        value_descriptor: Descriptor = None,
        **kwargs,
    ):
        """Swaps out the ParameterEH for a Parameter and PathSelector for backwards compatibility.

        :param parameter_path: the path to the parameter to modify for the EnergyHub
        :param value_descriptor: a Descriptor specifying which values to use
        :return a Parameter containing the appropriate PathSelector and descriptor.
        """
        selector = PathSelector(parameter_path)
        return Parameter(
            selector=selector, value_descriptors=value_descriptor, **kwargs
        )


class wwrSelector(Selector):
    """
    :class:`Selector <IO_Objects.Selector>` for window to wall ratio.
    """

    def __init__(self):
        super().__init__()

    def set(self, building, value):
        """Sets the window to wall ratio of the building to the provided value

        :param building: the building to modify
        :param value: the value of the wwr
        """
        ef.wwr_all(building, value)

    def get(self, building):
        # This feature has not yet been requested
        raise NotImplementedError(
            "Calculation of window to wall ratio is not supported."
        )

    def setup(self, building) -> None:
        """Adjusts the building to have at most one window per wall,
        making it ready to have its window to wall ratio modified.

        :param building: The building to adjust
        """
        ef.one_window(building)


def wwr(value_descriptor=None, **kwargs) -> Parameter:
    """Makes a window-to-wall-ratio parameter.

    :param value_descriptor: a parameter describing the valid window-wall-ratios
        Defaults to RangeParameter(0.01, 0.99, name="Window to Wall Ratio")
    :return: Parameter
    """
    value_descriptor = value_descriptor or RangeParameter(
        0.01, 0.99, name="Window to Wall Ratio"
    )
    if isinstance(value_descriptor, RangeParameter):
        min_val, max_val = value_descriptor.min, value_descriptor.max
        if not (0 < min_val < max_val < 1):
            if 0 == min_val:
                raise ValueError("min must be strictly greater than 0")
            if 1 == max_val:
                raise ValueError("max must be strictly less than 1")
            raise ValueError(
                "Invalid min and max values. 0 < min < max < 1 must be satisfied."
            )
    else:
        warnings.warn(
            f"wwr is intended to be used with RangeParameter. Your value_descriptor is {value_descriptor}"
        )

    selector = wwrSelector()
    return Parameter(selector=selector, value_descriptors=value_descriptor, **kwargs)


# Simpler ways to get specific transformers
keyFormat = Dict[str, Dict[str, Tuple[float, float]]]


# TODO: Formalize this syntax, maybe add a guaranteed order
def expand_plist(pList: keyFormat) -> List:
    """This function expands a nested dictionary of the correct format into a list of inputs.

    The dictionary should have the format:
    {'idf object name': {'idf object1 property name': (min_value, max_value)}

    Both layers of the dictionaries can have as many names as desired"""
    return [
        Parameter(
            FieldSelector(object_name=name, field_name=prop),
            RangeParameter(min_val=min_, max_val=max_, name=prop),
        )
        for name, subProps in pList.items()
        for prop, (min_, max_) in subProps.items()
    ]



from io import BytesIO
from pathlib import Path

import pytest
import warnings

from besos import config
from besos import eplus_funcs as eplus
from besos import eppy_funcs as ef
from besos.errors import InstallationError


@pytest.fixture(params=["idf", "json"])
def building(request):
    return ef.get_building(mode=request.param)


def test_get_ep_path():
    # first test regular get_ep_path
    ep_exe, ep_path = eplus.get_ep_path("9.0")
    # test ep_path works
    eplus.get_ep_path("9.0", ep_path=ep_path)
    # test error handling
    try:
        eplus.get_ep_path("9.3", ep_path=ep_path)
        raise ValueError(
            "Error should have been thrown as ep_path did not match ep_version"
        )
    except InstallationError:
        pass
    try:
        eplus.get_ep_path("9.0", ep_path="nonExistingDir")
        raise ValueError("ep_path does not exist, should have raised error")
    except InstallationError:
        pass


def test_get_possible_outputs(building, regtest):
    # NOTE: test relies on same idf to output same possible outputs.
    # test will most likely fail when new idfs are placed in config
    with regtest:
        eplus.print_available_outputs(building, name="Electricity:Facility")
        eplus.print_available_outputs(building, frequency="Monthly")



"""Enables some geomeppy based functionality on idf objects"""

# External Libraries
from geomeppy import view_geometry, recipes, IDF as geomeppy_idf
from numpy import sin, cos, deg2rad
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

try:
    from mpl_toolkits.mplot3d import Axes3D  # noqa
    import matplotlib.pyplot as plt
except (ImportError, RuntimeError):
    # this isn't always needed so we can ignore if it's not present
    pass


class IDF(geomeppy_idf):
    """Adding some class functions to help with Geomeppy"""

    def read(self):
        """don't need to use allcaps to search for items in idfobjects."""

        class AllCapsDict(dict):
            def __getitem__(self, key):
                return super().__getitem__(key.upper())

        super().read()
        self.idfobjects = AllCapsDict(self.idfobjects)

    def relative_to_world_coords(self):
        """translates all surfaces, subsurfaces, shadings and daylighting reference points
        to world coordinates"""
        relative_to_world_coords(self)

    def remove_windows(self, orientation=None):
        """removes windows with respect to orientation
        Note: default is ALL directions
        """
        remove_windows(self, orientation)

    def add_overhangs(self, depth, tilt=90, orientation=None, transmittance=""):
        """Set the overhang shading on all external windows by creating a SHADING:ZONE:DETAILED object
        which can be viewed using geomeppy's view_model() function.
        Note:
        Depth must be greater than 0
        :param depth: The depth of the overhang {m}.
        :param tilt: Tilt Angle from Window/Door {deg}.
        :param orientation: One of "north", "east", "south", "west". Walls within 45 degrees will be affected.
        :param Transmittance: Transmittance Schedule Name.
        """
        add_overhangs(self, depth, tilt, orientation, transmittance)

    def remove_shading(self, orientation=None):
        """removes shading surfaces with respect to orientation
        Note: default is ALL directions
        """
        remove_shading(self, orientation)

    # def translate_to_origin(self):
    #     """Translates building so it is touching the origin
    #     Note: overrides Geomeppy's translate_to_origin()
    #     """
    #     translate_to_origin(self)

    def view_model(self):
        """automatically checks if the idf is in Relative or World coordinates"""
        ggr = self.idfobjects["GLOBALGEOMETRYRULES"][0]
        if ggr.Coordinate_System == "Relative":
            return relative_coordinates_model(self)
        return view_idf(self)


def view_idf(idf):
    # type: (IDF) -> None
    """Display an IDF for inspection."""
    fig = plt.figure()

    # create the figure and add the surfaces
    ax = fig.add_subplot(projection="3d")
    collections = view_geometry._get_collections(idf, opacity=0.5)
    for c in collections:
        ax.add_collection3d(c)

    # calculate and set the axis limits
    limits = view_geometry._get_limits(idf=idf)
    ax.set_xlim(limits["x"])
    ax.set_ylim(limits["y"])
    ax.set_zlim(limits["z"])

    fig = plt.gcf()
    return fig, ax


def remove_shading(idf, orientation=None):

    """removes fins and overhangs from the idf
    Note:
    cannot differentiate between fins and overhangs.

    :param: orientation: compass direction with respect to building azimuth

    """

    from eppy.bunch_subclass import BadEPFieldError

    # orientation to degrees
    orientations = {
        "north": 0.0,
        "east": 90.0,
        "south": 180.0,
        "west": 270.0,
        None: None,
    }

    degrees = orientations.get(orientation, None)
    external_walls = filter(
        lambda x: x.Outside_Boundary_Condition.lower() == "outdoors",
        idf.getsurfaces("wall"),
    )

    external_walls = list(
        filter(lambda x: recipes._has_correct_orientation(x, degrees), external_walls)
    )
    windows = idf.getsubsurfaces("window")
    windows_walls = []
    for wall in external_walls:
        for window in windows:
            if window.Building_Surface_Name == wall.Name:
                windows_walls.append(window)
                windows_walls.append(wall)

    shading = idf.getshadingsurfaces()

    to_remove = [
        "Shading:Site:Detailed",
        "Shading:Building:Detailed",
    ]  # objects mentioned do not contain fins and overhangs
    for shade in list(shading):
        if shade.obj[0] in to_remove:
            shading.remove(shade)

    for surface in windows_walls:
        for shade in list(shading):

            try:
                if shade.Base_Surface_Name == surface.Name:
                    idf.removeidfobject(shade)
                    shading.remove(shade)
            except BadEPFieldError:
                if shade.Window_or_Door_Name == surface.Name:
                    idf.removeidfobject(shade)
                    shading.remove(shade)
            except BadEPFieldError:
                raise BadEPFieldError(
                    "unable to find field Base_Surface_Name and Window_or_Door_Name"
                )


def relative_to_world_coords(idf):

    """translates all surfaces, subsurfaces and daylighting reference points to world coordinates and
    then sets global geometry rules coordinate systems to "World"
    """
    # requires geomeppy.recipes.translate_coords
    from tqdm.notebook import tqdm

    shading = idf.idfobjects["SHADING:ZONE:DETAILED"]
    windows = idf.getsubsurfaces("window")
    surfaces = idf.getsurfaces()
    zones = idf.idfobjects["ZONE"]
    daylighting_refs = idf.idfobjects["DAYLIGHTING:REFERENCEPOINT"]

    for zone in tqdm(zones, desc="Relative to World Coordinates"):
        x, y, z = zone.X_Origin, zone.Y_Origin, zone.Z_Origin
        if zone.Direction_of_Relative_North == 180:
            x, y, z = -zone.X_Origin, -zone.Y_Origin, zone.Z_Origin
        zone_coords = (x, y, z)

        for surface in surfaces:
            coords = surface.coords
            if surface.Zone_Name == zone.Name:
                new_coords = recipes.translate_coords(coords, zone_coords)

                for window in windows:
                    if window.Building_Surface_Name == surface.Name:
                        win_coords = recipes.translate_coords(
                            window.coords, zone_coords
                        )
                        window.setcoords(win_coords)

                for shade in shading:
                    if shade.Base_Surface_Name == surface.Name:
                        shade_coords = recipes.translate_coords(
                            shade.coords, zone_coords
                        )
                        shade.setcoords(shade_coords)
                surface.setcoords(new_coords)

        for point in daylighting_refs:
            if point.Zone_Name == zone.Name:
                point.XCoordinate_of_Reference_Point += zone.X_Origin
                point.YCoordinate_of_Reference_Point += zone.Y_Origin
                point.ZCoordinate_of_Reference_Point += zone.Z_Origin

    ggr = idf.idfobjects["GLOBALGEOMETRYRULES"][0]
    ggr.Coordinate_System = "World"
    ggr.Daylighting_Reference_Point_Coordinate_System = "World"
    ggr.Rectangular_Surface_Coordinate_System = "World"


def add_overhangs(idf, depth=float, tilt=90, orientation=None, Transmittance=""):
    # uses geomeppy.recipes._has_correct_orientation

    """Set the overhang shading on all external windows by creating a SHADING:ZONE:DETAILED object
        which can be viewed using geomeppy's view_model() function.

    Note:
    Depth must be greater than 0

    :param idf: The IDF to edit.
    :param depth: The depth of the overhang {m}.
    :param tilt: Tilt Angle from Window/Door {deg}.
    :param orientation: One of "north", "east", "south", "west". Walls within 45 degrees will be affected.
    :param Transmittance: Transmittance Schedule Name.

    """

    try:
        ggr = idf.idfobjects["GLOBALGEOMETRYRULES"][0]  # type: Optional[Idf_MSequence]
    except IndexError:
        ggr = None
    # orientation to degrees
    orientations = {
        "north": 0.0,
        "east": 90.0,
        "south": 180.0,
        "west": 270.0,
        None: None,
    }
    degrees = orientations.get(orientation, None)
    external_walls = filter(
        lambda x: x.Outside_Boundary_Condition.lower() == "outdoors",
        idf.getsurfaces("wall"),
    )
    external_walls = list(
        filter(lambda x: recipes._has_correct_orientation(x, degrees), external_walls)
    )
    windows = idf.getsubsurfaces("window")
    for wall in external_walls:

        for window in windows:

            if window.Building_Surface_Name == wall.Name:
                coords = [window.coords[0]]
                x, y, z = window.coords[1]
                coords.append(
                    (
                        x + depth * sin(deg2rad(window.azimuth)) * sin(deg2rad(tilt)),
                        y + depth * cos(deg2rad(window.azimuth)) * sin(deg2rad(tilt)),
                        window.coords[0][2] - depth * cos(deg2rad(tilt)),
                    )
                )
                x, y, z = window.coords[2]
                coords.append(
                    (
                        x + depth * sin(deg2rad(window.azimuth)) * sin(deg2rad(tilt)),
                        y + depth * cos(deg2rad(window.azimuth)) * sin(deg2rad(tilt)),
                        window.coords[3][2] - depth * cos(deg2rad(tilt)),
                    )
                )
                coords.append(window.coords[3])
                Shade = idf.newidfobject(
                    "SHADING:ZONE:DETAILED",
                    Name="%s - Overhang" % window.Name,
                    Transmittance_Schedule_Name=Transmittance,
                    Base_Surface_Name=window.Building_Surface_Name,
                )
                Shade.setcoords(coords, ggr)


def translate_to_origin(idf):
    # type: (IDF) -> None

    """Move an IDF so the building touches the origin.
    :param idf: The IDF to edit.

    """
    from geomeppy.geom.polygons import Polygon3D

    surfaces = idf.getsurfaces()
    daylighting_refs = idf.idfobjects["DAYLIGHTING:REFERENCEPOINT"]
    zones = idf.idfobjects["ZONE"]

    min_y = min(min(Polygon3D(s.coords).ys) for s in surfaces)
    surfaces = filter(lambda x: min(Polygon3D(x.coords).ys) == min_y, surfaces)
    min_x = min(min(Polygon3D(s.coords).xs) for s in surfaces)

    idf.translate((-min_x, -min_y))

    for point in daylighting_refs:
        point.XCoordinate_of_Reference_Point -= min_x
        point.YCoordinate_of_Reference_Point -= min_y
    for zone in zones:
        if zone.Direction_of_Relative_North == 0:
            zone.X_Origin -= min_x
            zone.Y_Origin -= min_y
        elif zone.Direction_of_Relative_North == 180:
            zone.X_Origin += min_x
            zone.Y_Origin += min_y

    for point in daylighting_refs:
        point.XCoordinate_of_Reference_Point -= min_x
        point.YCoordinate_of_Reference_Point -= min_y
    for zone in zones:
        if zone.Direction_of_Relative_North == 0:
            zone.X_Origin -= min_x
            zone.Y_Origin -= min_y
        elif zone.Direction_of_Relative_North == 180:
            zone.X_Origin += min_x
            zone.Y_Origin += min_y


def relative_coordinates_model(building):
    # uses geomeppy.recipes.translate_coords()

    """translates RELATIVE coordinates of building surfaces to be viewed
        using view_model() function without altering original idf.

    Note:
    Does not correctly display after using geomeppy's rotate() function

    :param: building: The building we want displayed with RELATIVE coordinates

    """
    from copy import deepcopy

    # creates a copy of the building so it doesn't edit original idf
    idf = deepcopy(building)
    relative_to_world_coords(idf)
    idf.translate_to_origin()
    return idf.view_model()


def remove_windows(idf, orientation=None):
    # uses geomeppy.recipes._has_correct_orientation

    """removes windows from the idf

    :param: orientation: compass direction with respect to building north

    """

    # orientation to degrees
    orientations = {
        "north": 0.0,
        "east": 90.0,
        "south": 180.0,
        "west": 270.0,
        None: None,
    }
    degrees = orientations.get(orientation, None)
    external_walls = filter(
        lambda x: x.Outside_Boundary_Condition.lower() == "outdoors",
        idf.getsurfaces("wall"),
    )
    external_walls = list(
        filter(lambda x: recipes._has_correct_orientation(x, degrees), external_walls)
    )
    windows = idf.idfobjects["FENESTRATIONSURFACE:DETAILED"]
    external_wall_names = [wall.Name for wall in external_walls]

    for window in list(windows):
        if window.Building_Surface_Name in external_wall_names:
            idf.removeidfobject(window)


"""Some epJSON model viewing"""


def _get_collections(building, opacity=0.5):
    """Set up 3D collections for each surface type."""
    if isinstance(building, dict):
        surfaces = building.get("BuildingSurface:Detailed", {})
        fenestrations = building.get("FenestrationSurface:Detailed", {})
        shading_surfaces = building.get("Shading:Zone:Detailed", {})
    else:
        surfaces = building["BuildingSurface:Detailed"]
        fenestrations = building["FenestrationSurface:Detailed"]
        shading_surfaces = building["Shading:Zone:Detailed"]
    # set up the collections
    walls = _get_collection("wall", surfaces, opacity, facecolor="lightyellow")
    floors = _get_collection("floor", surfaces, opacity, facecolor="dimgray")
    roofs = _get_collection("roof", surfaces, opacity, facecolor="firebrick")
    shading = _get_collection(
        "shading", shading_surfaces, opacity, facecolor="darkolivegreen"
    )
    windows = _get_windows("window", fenestrations, opacity, facecolor="cornflowerblue")

    return walls, roofs, floors, shading, windows


def _get_collection(surface_type, surfaces, opacity, facecolor, edgecolors="black"):
    """Make collections from a list of EnergyPlus surfaces."""
    coords = []
    for surface in surfaces:
        if surface_type == "shading":
            surface_coords = []
            for v_name in surfaces[surface]["vertices"]:
                vertices = list(point for vertex, point in v_name.items())
                surface_coords.append(tuple(vertices))
            coords.append(surface_coords)
        elif (
            surfaces[surface]["surface_type"]
            and surfaces[surface]["surface_type"].lower() == surface_type.lower()
        ):
            surface_coords = []
            for v_name in surfaces[surface]["vertices"]:
                vertices = list(point for vertex, point in v_name.items())
                surface_coords.append(tuple(vertices))
            coords.append(surface_coords)

    trimmed_coords = [c for c in coords if c]  # dump any empty surfaces
    collection = Poly3DCollection(
        trimmed_coords, alpha=opacity, facecolor=facecolor, edgecolors=edgecolors
    )
    return collection


def _get_windows(surface_type, fenestrations, opacity, facecolor, edgecolors="black"):
    window_coords = []
    for window in fenestrations:
        vertex_1 = list(
            fenestrations[window][key]
            for key in fenestrations[window]
            if "vertex_1" in key
        )
        vertex_2 = list(
            fenestrations[window][key]
            for key in fenestrations[window]
            if "vertex_2" in key
        )
        vertex_3 = list(
            fenestrations[window][key]
            for key in fenestrations[window]
            if "vertex_3" in key
        )
        vertex_4 = list(
            fenestrations[window][key]
            for key in fenestrations[window]
            if "vertex_4" in key
        )
        window_coords.append(
            [tuple(vertex_1), tuple(vertex_2), tuple(vertex_3), tuple(vertex_4)]
        )
    trimmed_coords = [c for c in window_coords if c]  # dump any empty surfaces
    collection = Poly3DCollection(
        trimmed_coords, alpha=opacity, facecolor=facecolor, edgecolors=edgecolors
    )
    return collection


def _get_limits(building):
    if isinstance(building, dict):
        surfaces = building.get("BuildingSurface:Detailed", {})
        surfaces.update(building.get("Shading:Zone:Detailed", {}))
    else:
        surfaces = building["BuildingSurface:Detailed"]
        surfaces.update(building["Shading:Zone:Detailed"])
    x_coords = []
    y_coords = []
    z_coords = []
    for surface in surfaces:
        vertices = surfaces[surface]["vertices"]
        for point in vertices:
            x_coords.append(point["vertex_x_coordinate"])
            y_coords.append(point["vertex_y_coordinate"])
            z_coords.append(point["vertex_z_coordinate"])
    delta = max(
        (max(x_coords) - min(x_coords)),
        (max(y_coords) - min(y_coords)),
        (max(z_coords) - min(z_coords)),
    )
    x_max = min(x_coords) + delta
    y_max = min(y_coords) + delta
    z_max = min(z_coords) + delta
    return {
        "x": (min(x_coords), x_max),
        "y": (min(y_coords), y_max),
        "z": (min(z_coords), z_max),
    }


def view_epJSON(building):
    fig = plt.figure()

    ax = fig.add_subplot(projection="3d")
    collections = _get_collections(building, opacity=0.5)
    for c in collections:
        ax.add_collection3d(c)

    limits = _get_limits(building)

    ax.set_xlim(limits["x"])
    ax.set_ylim(limits["y"])
    ax.set_zlim(limits["z"])

    return fig, ax



import pytest

from besos import eppy_funcs as ef
from besos import sampling
from besos.evaluator import EvaluatorGeneric
from besos.parameters import (
    CategoryParameter,
    FieldSelector,
    Parameter,
    RangeParameter,
    wwr,
)
from besos.problem import Problem


# parameter - selector & descriptor
# selector - FilterSelector FieldSelector GenericSelector
# descriptor - RangeParameter CategoryParameter

# init, setup_checks, and setup_changes were moved directly from Will's code in
# parameters.py


@pytest.fixture(params=["idf", "json"])
def building(request):
    return ef.get_building(mode=request.param)


def test_init():
    # inputs should be initialisable
    Parameter(
        FieldSelector(object_name="NonRes Fixed Assembly Window", field_name="UFactor"),
        value_descriptors=RangeParameter(min_val=0.1, max_val=5),
    )
    wwr()
    wwr(value_descriptor=RangeParameter(0.01, 0.99, name="other"))
    print("Parameters initialised")


def test_load_idf():
    idf = ef.get_idf()
    assert len(idf.idfobjects["FENESTRATIONSURFACE:DETAILED"]) == 21, "bad idf"


@pytest.mark.xfail
def test_setup_checks():
    # try excepts were updated to asserts and now the test fails, no exceptions are
    # being raised and it seems to be in 3rd party code

    idf = ef.get_idf()

    # eppy key-inputs should reject bad key values at setup
    r: Parameter = Parameter(FieldSelector(class_name="invalid", field_name="any"))

    with pytest.raises(Exception):
        r.setup(idf)
    print("Invalid object key detected successfully.")
    r = Parameter(
        FieldSelector(object_name="NonRes Fixed Assembly Window", field_name="invalid")
    )
    with pytest.raises(Exception):
        r.setup(idf)
    print("Invalid property detected successfully.")


def test_setup_changes():
    idf = ef.get_idf()

    r1 = Parameter(
        FieldSelector(object_name="NonRes Fixed Assembly Window", field_name="UFactor"),
        value_descriptors=RangeParameter(min_val=0.1, max_val=5),
    )
    r2 = wwr()
    r3 = wwr(value_descriptor=RangeParameter(0.01, 0.99, name="other"))

    r1.setup(idf)
    r2.setup(idf)
    assert len(idf.idfobjects["FENESTRATIONSURFACE:DETAILED"]) == 4, "bug in r2.setup"

    idf = ef.get_idf()
    assert len(idf.idfobjects["FENESTRATIONSURFACE:DETAILED"]) == 21, "bad idf"
    r3.setup(idf)
    assert len(idf.idfobjects["FENESTRATIONSURFACE:DETAILED"]) == 4, "bug in r3.setup"

    print("setup tests done")


def test_custom_evaluation(regtest):
    """check to see if descriptors display as intended, and check to make sure custom
    evaluations work with EvaluatorGeneric"""

    # create the descriptors
    zero_to_nine = RangeParameter(min_val=0, max_val=9, name="0-9")
    single_digit_integers = CategoryParameter(
        options=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], name="single digit"
    )
    text_example = CategoryParameter(options=["a", "b", "c", "other"], name="text")

    # create the parameters and the problem
    parameters = [
        Parameter(value_descriptors=zero_to_nine),
        Parameter(value_descriptors=single_digit_integers),
        Parameter(value_descriptors=text_example),
    ]
    problem = Problem(parameters, outputs=["output"])

    # create the sampling distribution (seeded with only one output to give consistent
    # outputs after evaluating)
    samples = sampling.dist_sampler(sampling.seeded_sampler, problem, num_samples=1)
    with regtest:
        print(samples)

    # custom evaluation function from the jupyter notebook
    def evaluation_function(values):
        x, y, z = values
        if z == "other":
            return (0,)
        else:
            return (x * y,)

    evaluator = EvaluatorGeneric(evaluation_function, problem)
    # The evaluator will use this objective by default
    outputs = evaluator.df_apply(samples, keep_input=True)
    result = outputs.iloc[0]["output"]
    with regtest:
        print(f"{result:.5E}")


@pytest.fixture
def range_zero_to_one():
    return RangeParameter(0, 1, name="zero to one")


@pytest.fixture
def range_zero_to_five():
    return RangeParameter(0, 5, name="zero to five")


@pytest.fixture
def range_unnamed():
    return RangeParameter(0, 5)


@pytest.fixture
def category_abc():
    return CategoryParameter(["A", "B", "C"], name="ABC")


@pytest.fixture
def multiple_descriptor_parameter(range_zero_to_one, category_abc):
    return Parameter(
        value_descriptors=[range_zero_to_one, category_abc, range_zero_to_one]
    )


def test_multiple_descriptor_names(multiple_descriptor_parameter):
    assert Problem([multiple_descriptor_parameter]).names() == ["zero to one", "ABC"]


def test_multiple_descriptor_sampling(multiple_descriptor_parameter):
    problem = Problem([multiple_descriptor_parameter])
    samples = sampling.dist_sampler(sampling.lhs, problem, 5)
    assert len(samples.columns) == 2


def test_problem_value_descriptors(range_zero_to_five, range_zero_to_one, category_abc):
    problem = Problem(
        [
            Parameter(value_descriptors=lst)
            for lst in [
                [range_zero_to_one, range_zero_to_five],
                [category_abc, range_zero_to_one],
                [category_abc],
            ]
        ]
    )
    assert problem.num_inputs == 3
    assert problem.value_descriptors == [
        range_zero_to_one,
        range_zero_to_five,
        category_abc,
    ]


@pytest.mark.compatibility
def test_parameter_name_transfers_to_descriptor(range_unnamed):
    with pytest.warns(
        DeprecationWarning,
        match=r"Call to deprecated function \(or staticmethod\) value_descriptor.",
    ):
        Parameter(value_descriptors=range_unnamed, name="foo")
    assert range_unnamed.name == "foo"


@pytest.mark.compatibility
def test_parameter_name_does_not_overwrite(range_zero_to_five):
    old_name = range_zero_to_five.name
    with pytest.warns(None) as record:
        Parameter(value_descriptors=range_zero_to_five, name="foo")
    assert (
        "This parameter's descriptor is already named"
        in record.pop(UserWarning).message.args[0]
    )
    assert (
        "Call to deprecated function (or staticmethod) value_descriptor."
        in record.pop(DeprecationWarning).message.args[0]
    )
    assert range_zero_to_five.name == old_name


def test_abstractfield_arithmetic(building):
    mode = ef.get_mode(building)
    if mode == "idf":
        index = None
        for obj in building.idfobjects[
            "MATERIAL"
        ]:  # for loop to find index of "Mass NonRes Wall Insulation"
            if obj.Name == "Mass NonRes Wall Insulation":
                index = building.idfobjects["MATERIAL"].index(obj)
                break
        start = float(
            building.idfobjects["MATERIAL"][index].Thickness
        )  # chose arbitrarly as it has thickness values to work with
        testSelector = FieldSelector(
            class_name="MATERIAL",
            object_name="Mass NonRes Wall Insulation",
            field_name="Thickness",
        )
        testSelector.multiply(building, 2)
        testSelector.add(building, -0.03)  # arbitrary numbers to check for arithmetic
        end = building.idfobjects["MATERIAL"][index].Thickness
    else:
        start = building["Material"]["Mass NonRes Wall Insulation"][
            "thickness"
        ]  # chose arbitrarly as it has thickness values to work with
        testSelector = FieldSelector(
            class_name="Material",
            object_name="Mass NonRes Wall Insulation",
            field_name="thickness",
        )
        testSelector.multiply(building, 2)
        testSelector.add(building, -0.03)
        end = building["Material"]["Mass NonRes Wall Insulation"]["thickness"]
    assert (start * 2) - 0.03 == end



"""
Functions related to under-the-hood interactions with energyplus.
"""

# Python Core Libraries
import json
import os
import platform
import shutil
import subprocess
import tempfile
import warnings
from pathlib import Path

# External Libraries
from deprecated.sphinx import deprecated

# BESOS Imports
from besos import eppy_funcs as ef
from besos import config
from besos import objectives
from besos.errors import ModeError, InstallationError
from besos.besostypes import PathLike


# TODO: Make this a method of the building class when we add it
def get_idf_version(building):
    """Get energyplus version from idf or json building
    :param building: IDF object to find version of.
    :returns: String of version (ex. '9.3.0')
    """
    mode = ef.get_mode(building)
    if mode == "idf":
        return building.idfobjects["VERSION"][0].Version_Identifier
    elif mode == "json":
        try:
            return building["Version"]["Version 1"]["version_identifier"]
        except KeyError:
            try:
                return building["epJSON_schema_version"]
            except:
                raise ModeError(
                    mode="json",
                    message=f"Cannot find IDF version, check epJSON formatting.",
                )
    else:
        raise ModeError(mode)


def get_idf_version_pth(path):
    """Get energyplus version from idf or epJSON path
    :param path: Path to idf to epJSON file.
    :returns: String of version (ex. '9.3.0')
    """
    path = str(path)
    version_lines = (
        ""  # string containing the important lines of the json file to later
    )
    version = ""
    flag = False  # used to save the lines following 'Version' to version_lines
    if path.endswith(".idf"):
        with open(path, "r") as read_obj:  # opening idf file to read
            # Read all lines in the file one by one
            for line in read_obj:
                if flag:
                    version_lines += line.rstrip(
                        "\n"
                    )  # adds the line after 'Version,' is seen in case of line break
                    break
                # For each line, check if line contains 'Version,'
                if ("Version," in line or "VERSION," in line) and not line.startswith(
                    "!"
                ):
                    version_lines += line.rstrip("\n")
                    flag = True
        version = (
            version_lines.split(";")[0].split(",")[-1].replace(" ", "").replace("=", "")
        )
    elif path.endswith(".epJSON"):
        readobj = open(path, "r")
        jsonText = readobj.read()
        readobj.close()
        jsonObj = json.loads(jsonText)
        try:
            version = jsonObj["Version"]["Version 1"]["version_identifier"]
        except KeyError:
            try:
                version = jsonObj["epJSON_schema_version"]
            except KeyError:
                message = f"Version object not found within epJSON file, consider updating epJSON file\nepJSON File: {path}"
                raise ModeError("json", message=message)
    return version


def has_hvac_templates(building) -> bool:
    """Returns whether or not the building contains HVACTemplate objects
    https://bigladdersoftware.com/epx/docs/8-0/input-output-reference/page-061.html

    :param building:
    :return: True if at leas one HVACTemplate object is present in the building.
    """
    mode = ef.get_mode(building)
    prefix = ef.convert_format("HVACTemplate", "class", mode)
    if mode == "idf":
        return any(
            k for k, v in building.idfobjects.items() if k.startswith(prefix) and v
        )
    else:
        return any(k for k in building if k.startswith(prefix))


def run_building(
    building,
    out_dir=config.out_dir,
    version=None,
    **eplus_args,
):
    """Run energy plus on a building object and return results
    :param building_path: Path to building file
    :param out_dir: Path to store EnergyPlus output files, if out_dir is not defined the results will not be saved
    :param version: Version of building file
    :param ep_path: Path to EnergyPlus if installed in unexpected directory
    :param expand_objects: Boolean if '--expandobjects' option to be appended to command
    :returns: A dictionary of EnergyPlus outputs
    """
    # backwards compatibility
    if version:
        warnings.warn(
            "the version argument is deprecated for run_building,"
            " and will be removed in the future",
            FutureWarning,
        )
        assert version == get_idf_version(building), "Incorrect version"

    with tempfile.TemporaryDirectory(dir=Path.home(), prefix=".besos_") as temp_dir:
        if out_dir is None:
            out_dir = temp_dir
        try:
            building_path = Path(temp_dir, "in.idf").resolve()
            building.saveas(str(building_path))
        except AttributeError:
            building_path = Path(temp_dir, "in.epJSON").resolve()
            with open(str(building_path), "w") as f:
                json.dump(building, f)
        expand_objects = has_hvac_templates(building)
        run_energyplus(
            building_path,
            out_dir=out_dir,
            version=get_idf_version(building),
            expand_objects=expand_objects,
            **eplus_args,
        )
        return objectives.read_eso(out_dir)


def run_energyplus(
    building_path: PathLike,
    out_dir: PathLike = config.out_dir,
    epw: PathLike = config.files["epw"],
    err_dir: PathLike = config.err_dir,
    schema_file=None,
    error_mode="Silent",
    version=None,
    ep_path=None,
    expand_objects: bool = False,
    *,
    stdout_mode="Silent",
):
    """Run EnergyPlus. This method is intended to work as similar to the cli tool as possible
    :param building_path: Path to building file
    :param out_dir: Path to store EnergyPlus output files
    :param epw: Path to epw file
    :param err_dir: Path to store EnergyPlus error files
    :param schema_file: Path to data dictionary for EnergyPlus
    :param stdout_mode: Stdout mode selection, options are "Verbose" and "Silent"
    :param error_mode: Error mode selection
    :param version: Version of building file
    :param ep_path: Path to EnergyPlus if installed in unexpected directory
    :param expand_objects: Boolean if '--expandobjects' option to be appended
    :param stdout_mode: Stdout mode selection. One of {"Silent", "Verbose"} raise warning otherwise.
        Silent: EnergyPlus stdout output is suppressed. This is the default.
        Verbose: EnergyPlus output is printed to stdout.
    :returns: None
    """

    if version is None:
        version = get_idf_version_pth(building_path)
    ep_exe_path, ep_directory = get_ep_path(version, ep_path)
    schema_file = schema_file or Path(ep_directory, "Energy+.idd")

    cmd = [
        ep_exe_path,
        "--idd",
        schema_file,
        "--weather",
        epw,
    ]
    if out_dir:
        cmd += ["--output-directory", out_dir]
    if expand_objects:
        cmd.append("--expandobjects")
    cmd.append(building_path)
    needs_shell = platform.system() == "Windows"

    # validate stdout_mode
    stdout_mode_options = {"Silent", "Verbose"}
    warning = (
        "Invalid value for stdout_mode. "
        'Must be one of {"Silent", "Verbose"}. '
        'Using default value of "Silent". '
    )
    stdout_arg = {"stdout": subprocess.DEVNULL}
    if stdout_mode not in stdout_mode_options:
        warnings.warn(warning)
    elif stdout_mode == "Verbose":
        stdout_arg = {}
    try:
        if stdout_arg:
            print("Running EnergyPlus with stdout output suppressed...")
        subprocess.run(cmd, check=True, shell=needs_shell, **stdout_arg)
    except subprocess.CalledProcessError as e:
        # TODO: This log is excessively noisy. Can we cut it down to just the command's stderr?
        if error_mode != "Silent":
            # print eplus error
            filename = Path(out_dir, "eplusout.err")
            if os.path.exists(filename):
                err_file = open(filename, "r")
                for line in err_file:
                    print(line)
                print()
                err_file.close()
        if err_dir is not None and out_dir != err_dir:
            # copy eplus error files to err_dir
            if os.path.exists(err_dir):
                shutil.rmtree(err_dir)
            shutil.copytree(out_dir, err_dir)
        raise e


def get_ep_path(idf_version, ep_path=None):
    """get EnergyPlus installation path by idf_version
    :param idf_version: idf_version of EnergyPlus
    :param ep_path: Optional - path to EnergyPlus if installed in different location then expected
    :returns: (path to EnergyPlus executable, path to EnergyPlus)
    """
    if "." in idf_version:
        if idf_version == "9.0" or idf_version == "9.0.1":
            idf_version = "9-0-1"
        else:
            idf_version = idf_version.replace(".", "-") + "-0"
    if ep_path is not None:
        ep_directory = os.path.abspath(ep_path)
        if not os.path.exists(ep_directory):
            message = f"'ep_path' does not exist.\n'ep_path': {ep_path}"
            raise InstallationError(message=message)
        try:
            cmd = [os.path.join(ep_directory, "energyplus"), "--version"]
            out, err = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()
            ep_path_version = (
                str(out).split("Version ")[-1].split("-")[0].replace(".", "-")
            )
        except:
            message = f"'ep_path' provided does not contain an energyplus file\n'ep_path': {ep_path}"
            raise InstallationError(message=message)
        if ep_path_version != idf_version:
            message = f"'ep_path' provided is a different version of EnergyPlus then 'idf_version' argument\n'ep_path': {ep_path}\n'idf_version': {idf_version}"
            raise InstallationError(message=message)
        if platform.system() == "Windows":
            ep_exe = os.path.join(ep_directory, "energyplus.exe")
        else:
            ep_exe = os.path.join(ep_directory, "energyplus")
    else:
        # this is duplicated from eppy.runner.run_functions.paths_from_version
        if platform.system() == "Windows":
            ep_directory = "C:/EnergyPlusV{idf_version}".format(idf_version=idf_version)
            ep_exe = os.path.join(ep_directory, "energyplus.exe")
        elif platform.system() == "Linux":
            ep_directory = "/usr/local/EnergyPlus-{idf_version}".format(
                idf_version=idf_version
            )
            ep_exe = os.path.join(ep_directory, "energyplus")
        else:
            ep_directory = "/Applications/EnergyPlus-{idf_version}".format(
                idf_version=idf_version
            )
            ep_exe = os.path.join(ep_directory, "energyplus")

    if not os.path.exists(ep_directory) or not os.path.exists(ep_exe):
        raise InstallationError(version=idf_version)
    return ep_exe, ep_directory


def print_available_outputs(
    building,
    version=None,
    name=None,
    frequency=None,
):
    # backwards compatibility
    if version:
        warnings.warn(
            "the version argument is deprecated for print_available_outputs,"
            " and will be removed in the future",
            FutureWarning,
        )
        assert version == get_idf_version(building), "Incorrect version"

    if name is not None:
        name = name.lower()
    if frequency is not None:
        frequency = frequency.lower()
    results = run_building(building, stdout_mode="Verbose")
    for key in results.keys():
        if name is not None:
            if name not in key[0].lower():
                continue
            if frequency is not None and key[1].lower() != frequency:
                continue
        elif frequency is not None:
            if key[1].lower() != frequency:
                continue
        print(list(key))



from io import BytesIO
from pathlib import Path
import warnings

from eppy.modeleditor import IDDAlreadySetError
import pytest

from besos import config
from besos import eplus_funcs
from besos import eppy_funcs as ef


@pytest.fixture(params=["idf", "json"])
def building(request):
    return ef.get_building(mode=request.param)


@pytest.fixture
def clean_idd_name():
    # reset the iddname
    # this kind of monkeypatch should not be used in production
    # it may mess with how idfs are processed

    # clear name before test
    old_name = ef.IDF.iddname
    ef.IDF.iddname = None
    yield
    # restore name after test
    ef.IDF.iddname = old_name


@pytest.fixture
def idd_other(clean_idd_name):
    return Path(config.data_dir, "Custom_Long_Fields.idd")


def test_idd_no_conflict_defaults(clean_idd_name):
    ef.get_idf()
    ef.get_idf()


def test_idd_no_conflict_same(idd_other, clean_idd_name):
    ef.get_idf(idd_file=idd_other)
    ef.get_idf(idd_file=idd_other)


def test_idd_conflict(idd_other, clean_idd_name):
    with pytest.raises(IDDAlreadySetError):
        ef.get_idf()
        ef.get_idf(idd_file=idd_other)


def test_idd_warns(idd_other, clean_idd_name):
    ef.get_idf(idd_file=idd_other)
    with pytest.warns(UserWarning, match="idd is already set to: "):
        ef.get_idf()


def test_view_building(building, regtest):
    """Verify that the view building function gives consistent results."""
    fig, ax = ef.view_building(building)
    out = BytesIO()
    fig.savefig(out)
    with regtest:
        print(out.getvalue())


def test_remove_shading():
    building = ef.get_building(mode="idf")
    building.add_overhangs(
        0.3
    )  # NOTE: Floats are completely arbitrary and can be edited
    building.remove_shading()
    assert building.getshadingsurfaces() == []


def test_remove_windows():
    building = ef.get_building(mode="idf")
    building.remove_windows()
    assert (
        len(building.idfobjects["FENESTRATIONSURFACE:DETAILED"]) == 0
    )  # test relies on all FENESTRATIONSURFACE:DETAILED objects being windows


def test_wwr_all(building, regtest):
    with warnings.catch_warnings():
        warnings.simplefilter(
            "ignore", category=UserWarning
        )  # ignores glassdoor warning
        ef.wwr_all(
            building, 0.235
        )  # NOTE: Floats and directions are completely arbitrary and can be edited
        no_direction = ef.get_windows(building)
        ef.wwr_all(building, 0.5, direction="north")
        with_direction = ef.get_windows(building)
    with regtest:
        print(no_direction)
        print(with_direction)


def test_set_daylight_control(regtest):
    building = ef.get_building(mode="idf")
    ef.set_daylight_control(building, zone_name="Perimeter_ZN_1", distance=2)
    with regtest:
        print(building.idfobjects["Daylighting:Controls"])



import random
import shutil

import platypus
import pytest

from besos import eppy_funcs as ef
from besos import optimizer
from besos import pyehub_funcs as pf
from besos.evaluator import EvaluatorEP, EvaluatorEH
from besos.parameters import (
    Parameter,
    RangeParameter,
    PathSelector,
    expand_plist,
)
from problem import EPProblem, EHProblem


needs_glpk = pytest.mark.skipif(
    shutil.which("glpsol") is None,
    reason=("Requires a PuLP compatible solver. Tests assume GLPK is the solver used."),
)


@pytest.fixture
def parameters():
    parameters = expand_plist(
        {
            "NonRes Fixed Assembly Window": {
                "UFactor": (0.1, 5),
                "Solar Heat Gain Coefficient": (0.01, 0.99),
            },
            "Mass NonRes Wall Insulation": {"Thickness": (0.01, 0.09)},
        }
    )
    return parameters


@pytest.fixture
def EHparameters():
    parameters = [
        Parameter(
            PathSelector(["LINEAR_CAPITAL_COSTS", "Boiler"]),
            value_descriptors=RangeParameter(100.0, 200.0, name="Boiler"),
        ),
        Parameter(
            PathSelector(["LINEAR_CAPITAL_COSTS", "CHP"]),
            value_descriptors=RangeParameter(300.0, 400.0, name="CHP"),
        ),
    ]
    return parameters


@pytest.fixture
def EHproblem(EHparameters):
    objectives = ["total_cost", "total_carbon"]
    return EHProblem(EHparameters, objectives)


@pytest.fixture
def problem(parameters):
    objectives = ["Electricity:Facility", "Gas:Facility"]
    return EPProblem(parameters, objectives)


@pytest.mark.slow
def test_flexibility(problem, regtest):
    """to make sure that you can use multiple algorithms on the same data set"""
    idf = ef.get_idf()
    evaluator = EvaluatorEP(problem, idf)

    random.seed(1)
    # run the first algorithm
    platypus_problem = evaluator.to_platypus()
    algorithm = platypus.NSGAII(problem=platypus_problem, population_size=5)
    algorithm.run(5)

    # run the second algorithm
    generator = platypus.InjectedPopulation(algorithm.population)
    alg2 = platypus.EpsMOEA(
        problem=platypus_problem, generator=generator, epsilons=3, population_size=5
    )
    alg2.run(5)

    results = optimizer.solutions_to_df(
        alg2.result, problem, parts=["inputs", "outputs"]
    )

    value = results.iloc[0]["Electricity:Facility"]
    with regtest:
        print(f"{value:.5E}")
    assert all(
        [optimal for optimal in results["pareto-optimal"]]
    ), f"Algorithm not producing optimal outputs"


@needs_glpk
@pytest.mark.slow
def test_energyhub(EHproblem):
    hub = pf.get_hub()
    evaluatorEH = EvaluatorEH(EHproblem, hub)
    platypus_problem = evaluatorEH.to_platypus()
    EHalgorithm = platypus.NSGAII(platypus_problem)
    EHalgorithm.run(5)
    results1 = optimizer.solutions_to_df(
        EHalgorithm.result, EHproblem, parts=["inputs", "outputs"]
    )
    best = results1.loc[results1["pareto-optimal"]]
    value = best.iloc[0]["total_cost"]

    assert value < 1000, f"Unexpected result: {value}"



"""Type definitions used for BESOS' type hints.

PathLike is defined a any object implementing ``os.pathlike`` or a string.
It is an alias for ``Union[os.PathLike, str]``
"""

# Python Core Libraries
import os
from typing import Union


PathLike = Union[os.PathLike, str]



import random

import pytest

from besos import optimizer
from besos.evaluator import EvaluatorGeneric
from besos.parameters import CategoryParameter, Parameter, RangeParameter
from besos.problem import Problem


@pytest.fixture
def parameters():
    parameters = [
        Parameter(value_descriptors=CategoryParameter(list(range(80, 90)), name="1st")),
        Parameter(value_descriptors=CategoryParameter(list(range(80, 90)), name="2nd")),
        Parameter(value_descriptors=RangeParameter(3, 100, name="3rd")),
    ]
    return parameters


def mixed_types(values):
    # helper function from the jupyter notebook
    values = list(values)
    num = values.pop(-1)
    return tuple(num % v for v in values)


def test_mixed_type_parameters(parameters, regtest):
    """to make sure that we can use different parameter types in the same algorithms
    while still getting optimal outputs"""

    evaluator = EvaluatorGeneric(mixed_types, Problem(parameters, 2))
    random.seed(1)
    results = optimizer.EpsMOEA(evaluator, epsilons=10)

    results = results[["outputs_0", "outputs_1"]].values[0]
    with regtest:
        print([f"{x:.5E}" for x in results])


def test_optimizer_customization(parameters, regtest):
    """testing the variator field while using mixed parameters"""

    evaluator = EvaluatorGeneric(mixed_types, Problem(parameters, 2))
    variator = optimizer.get_operator(evaluator.to_platypus())

    random.seed(2)
    results = optimizer.EpsMOEA(evaluator, variator=variator, epsilons=10)

    results = results[["outputs_0", "outputs_1"]].values[0]
    with regtest:
        print([f"{x:.5E}" for x in results])



import pytest

from besos import eppy_funcs as ef
from besos import sampling
from besos.evaluator import EvaluatorEP
from besos.objectives import MeterReader, clear_outputs
from besos.optimizer import NSGAII
from besos.parameters import expand_plist
from besos.problem import EPProblem


@pytest.fixture(params=["idf", "json"])
def building(request):
    return ef.get_building(mode=request.param)


@pytest.fixture
def parameters():
    # The parameters for the problem/evaluator
    parameters = expand_plist(
        {
            "NonRes Fixed Assembly Window": {
                "U-Factor": (0.1, 5),
                "Solar Heat Gain Coefficient": (0.01, 0.99),
            },
            "Mass NonRes Wall Insulation": {"Thickness": (0.01, 0.09)},
        }
    )
    return parameters


@pytest.mark.slow
def test_objectives(building, parameters, regtest):
    """Testing custom functions and basic objective creation"""

    def variance(result):
        return result.data["Value"].var()

    objectives = [
        MeterReader("Electricity:Facility", name="Electricity Usage"),
        MeterReader("Electricity:Facility", func=variance, name="Electricity Variance"),
    ]
    problem = EPProblem(inputs=parameters, outputs=objectives)

    evaluator = EvaluatorEP(problem, building)
    samples = sampling.dist_sampler(sampling.seeded_sampler, problem, 10)
    results = evaluator.df_apply(samples, keep_input=True)

    value = results.iloc[0]["Electricity Variance"]

    with regtest:
        print(f"{value:.5E}")


@pytest.mark.slow
def test_constraints(building, parameters):
    """Testing for expected output with certain constraints, also acts as a test for using NSGAII"""

    objectives = ["Electricity:Facility", "Gas:Facility"]
    problem = EPProblem(
        inputs=parameters,
        outputs=objectives,
        constraints=["CO2:Facility"],
        constraint_bounds=[">=750"],
    )

    evaluator = EvaluatorEP(problem, building)

    results = NSGAII(evaluator, evaluations=10, population_size=2)

    value = results.iloc[0]["CO2:Facility"] + results.iloc[0]["violation"]
    assert (
        value >= 750
    ), f"Constraint did not effect output, value should be above 750 but was: {value}"

    # Check to make sure the output changes with different constraints
    problem = EPProblem(
        inputs=parameters,
        outputs=objectives,
        constraints=["CO2:Facility"],
        constraint_bounds=["<=750"],
    )

    evaluator = EvaluatorEP(problem, building)

    results = NSGAII(evaluator, evaluations=10, population_size=2)

    value = results.iloc[0]["CO2:Facility"] - results.iloc[0]["violation"]
    assert (
        value <= 750
    ), f"Constraint did not effect output, value should be below 750 but was: {value}"


def test_clear_outputs(building):
    mode = ef.get_mode(building)
    if mode == "idf":
        clear_outputs(building, outputs="OUTPUT:TABLE:MONTHLY")
        assert building.idfobjects["OUTPUT:TABLE:MONTHLY"] == []
    else:
        clear_outputs(building, outputs="Output:Table:Monthly")
        assert building["Output:Table:Monthly"] == {}
